
import logging
import re
from copy import deepcopy

from namematch.utils.utils import *

logger = logging.getLogger()


class Variable():

    def __init__(self, validated_variable_dict):

        for attribute, attribute_value in validated_variable_dict.items():
            setattr(self, attribute, attribute_value)


    @classmethod
    def build(cls, variable_dict, params):
        '''Create a Variable instance.

        Args:
            variable_dict (dict): info about a variable definition from user-input config
            params (dict): dictionary with processed match parameter values

        Returns:
            :mod:`namematch.data_structures.variable.Variable`: instance of the Variable class
        '''

        validated_variable_dict = dict()
        validated_variable_dict['check'] = ''
        validated_variable_dict['drop_list'] = []
        validated_variable_dict['colname_dict'] = {}
        validated_variable_dict['set_missing_list'] = []

        if 'name' not in variable_dict or 'compare_type' not in variable_dict:
            logger.error(f"Error in variable definition. Please ensure "
                         f"name and compare_type are specified.")
            raise

        validated_variable_dict['name'] = variable_dict['name']
        validated_variable_dict['compare_type'] = variable_dict['compare_type']
        is_last_name = (variable_dict["name"] == params.last_name_column)
        if is_last_name and params.split_names:
            validated_variable_dict["compare_type"] = "LastName"

        col_var_list = [k for k in variable_dict.keys() if '_col' in k]
        for col_var in col_var_list:
            file_nickname = re.sub('_col$', '', col_var)
            validated_variable_dict['colname_dict'][file_nickname] = variable_dict[col_var]
            # NOTE: will validate in NameMatch.build next step

        # issue warning if user specifies param not understood by code
        params_user_specified = [v for v in list(variable_dict.keys()) if '_col' not in v]
        params_acceptable = ['name', 'compare_type', 'check', 'drop', 'set_missing']
        params_unrecognized = [
                param for param in params_user_specified
                if param not in params_acceptable]
        if len(params_unrecognized) > 0:
            params_unrecognized = ', '.join(params_unrecognized)
            logger.warning(f"The following parameters of {validated_variable_dict['name']} "
                           f"are not recognized (will be ignored): {params_unrecognized}.")

        if 'check' in variable_dict:
            validated_variable_dict['check'] = variable_dict['check']

        if 'drop' in variable_dict:
            if not isinstance(variable_dict['drop'], list):
                logger.error(f"The drop parameters of {validated_variable_dict['name']}"
                             f"must be a list. It's currently a {type(variable_dict['drop'])}.")
                raise Exception
            validated_variable_dict['drop_list'] = variable_dict['drop']

        if 'set_missing' in variable_dict:
            if not isinstance(variable_dict['set_missing'], list):
                logger.error(f"The set_missing parameters of {validated_variable_dict['name']}"
                             f"must be a list. It's currently a {type(variable_dict['set_missing'])}.")
                raise Exception
            validated_variable_dict['set_missing_list'] = variable_dict['set_missing']

        return cls(validated_variable_dict)


    def validate_col_parameters(self, data_files):
        '''Validate that each data file has a corresponding "_col" parameter in each
        variable defintion.

        Args:
            data_files (:mod:`namematch.data_structures.data_file.DataFileList`): info about what input files are being matched
        '''

        col_var_name_list = list(self.colname_dict.keys())

        data_file_name_list = data_files.get_all_nicknames()

        missing_col_var_list = []
        for data_file_name in data_file_name_list:
            if data_file_name not in col_var_name_list:
                missing_col_var_list.append(data_file_name)

        if len(missing_col_var_list) > 0:
            missing_col_var_list = ', '.join(missing_col_var_list)
            logger.error(f'The following data files do not have a column definition '
                         f'in the {self.name} variable: {missing_col_var_list}')
            raise ValueError

        if len(col_var_name_list) > len(data_file_name_list):
            logger.warning(f'There are more column definitions in the {self.name} variable ',
                            f'than data files. Column definitions that do not correspond to '
                            f'a data file will be ignored.')


    def get_columns_to_read(self, file_nickname):
        '''Get the name(s) of the column(s) from the input file that need to be read in order to
        create the current all-names column.

        Args:
            file_nickname (str): nickname of input file being searched

        Return:
            list of column names
        '''

        column_name_list = self.colname_dict[file_nickname].split(',')

        all_column_names = []
        for column_name in column_name_list:
            if ',' in column_name:  # geography
                all_column_names.extend(column_name.split(','))
            else:
                if column_name != '':
                    all_column_names.append(column_name)

        return all_column_names


    def get_an_columns(self):
        '''Get the name(s) of the current all-names column(s).

        Return:
            list of column names
        '''

        if self.compare_type == 'Address':
            return ['address_street_number', 'address_street_name', 'address_street_type']
        else:
            return self.name


    def copy(self):
        '''Create a deep copy of a Variable object.'''
        return deepcopy(self)



class VariableList():
    '''Class that houses a list of Variable objects.'''

    def __init__(self, variable_list):

        self.varlist = variable_list

    @classmethod
    def build(cls, variable_dict_list, params):
        '''Create a VariableList instance.

        Args:
            variable_dict_list (dict): dictionary with variable info from user-input config
            params (dict): dictionary with processed match parameter values

        Returns:
            :mod:`namematch.data_structures.variable.VariableList`: instance of the VariableList class
        '''

        validated_variable_list = []
        for variable_dict in variable_dict_list:
            validated_variable = Variable.build(variable_dict, params)
            validated_variable_list.append(validated_variable)

        return cls(validated_variable_list)

    @classmethod
    def load(cls, variables_list_dict):
        '''Load a VariableList instance.

        Args:
            variables_list_dict (dict): dictionary version of a VariableList instance

        Returns:
            :mod:`namematch.data_structures.variable.VariableList`: instance of the VariableList class
        '''

        variable_list = [
                Variable(variable_dict)
                for variable_dict in variables_list_dict['varlist']]

        # NOTE: no need for validation because simple loading
        # an object that was validated when initialized

        return cls(variable_list)

    def validate_col_parameters(self, data_files):
        '''Validate that the "_col" variables referenced in the config's variable definitions
        actually exist in the input datasets.

        Args:
            data_files (:mod:`namematch.data_structures.data_file.DataFileList`): info about what input files are being matched
        '''

        for variable in self.varlist:
            variable.validate_col_parameters(data_files)

        # make sure the variables actually exist
        for data_file in (data_files.data_files + data_files.existing_data_files):
            df = pd.read_csv(data_file.filepath, nrows=1)
            cols_to_read = self.get_columns_to_read(data_file)
            for col in cols_to_read:
                if col not in df.columns.tolist():
                    logger.error(f"The {col} column does not exist in the "
                                 f"{data_file.nickname} file.")
                    raise

    def validate_variable_names(self):
        '''Validate that the Variables in the VariableList all have unique nicknames.'''

        # make sure names unique
        variable_name_list = self.get_names()
        if len(variable_name_list) != len(set(variable_name_list)):
            logger.error('Variables must have unique names.')
            raise ValueError

        # if variable with compare_type Adress exists, check to see if variables
        # address_street_number, address_street_name, or address_street_type already exist
        if any([variable.compare_type == 'Address' for variable in self.varlist]):
            addr_vars = ['address_street_number', 'address_street_name', 'address_street_type']
            variable_names_to_check = addr_vars
            variable_names_invalid = []
            for variable_name in variable_names_to_check:
                if variable_name in variable_name_list:
                    variable_names_invalid.append(variable_name)
            if len(variable_names_invalid) > 0:
                variable_names_invalid = ', '.join(variable_names_invalid)
                logger.error(f'Given a variable of compare_type Address, the following '
                             f'variable names are invalid: {variable_names_invalid}')
                raise ValueError

    def validate_type_counts(self, incremental):
        '''Validate that there is exactly one variable with compare type UniqueID. If
        incremental, validate that there is exactly one variable with compare type ExistingID.

        Args:
            incremental (bool): True if the config file provides "existing" data files
        '''

        # count number of unique ids and issue warning if greater than 1
        n_uids = len(self.get_variables_where(attr='compare_type', attr_value='UniqueID'))
        if n_uids > 1:
            logger.warning(f'You have more than one variable with compare_type '
                           f'UniqueID. This is only recommended in rare cases -- '
                           f'please review the documentation.')

        # if count number of existing ids and raise error if greater than 1
        if incremental:
            n_eids = len(self.get_variables_where(attr='compare_type', attr_value='ExistingID'))
            if n_eids == 0:
                logger.error(f'You must specify a variable with compare type ExistingID for incremental '
                             f'matching, -- please review the documentation.')
                raise ValueError
            if n_eids > 1:
                logger.error(f'You have more than one variable with compare_type ExistingID, '
                             f'which is prohibited -- please review the documentation.')
                raise ValueError

    def validate(self, data_files):
        '''Validate several components of the variables defined in the config.

        Args:
            data_files (:mod:`namematch.data_structures.data_file.DataFileList`): info about what input files are being matched
        '''

        self.validate_col_parameters(data_files)
        self.validate_variable_names()
        self.validate_type_counts(incremental=(len(data_files.existing_data_files) > 0))

    def get_variables_where(self, attr, attr_value, equality_type='equals', return_type='name'):
        '''Select variables that meet a certain condition (e.g. compare_type == 'Category').

        Args:
            attr (str): variable feature to condition on
            attr_value (str): acceptable values for the variable feature
            equality_type (str): check conditions using either "equals" or "in"
            return_type (str): either "name" or "ix", determining what return type to use

        Return:
            list of variable nicknames (all-names columns) or corresponding all-names column indices
        '''

        ix_list = []
        name_list = []
        for ix, variable in enumerate(self.varlist):
            if equality_type == 'equals':
                if variable.__dict__[attr] == attr_value:
                    ix_list.append(ix)
                    name_list.append(variable.name)
            elif equality_type == 'in':
                if variable.__dict__[attr] in attr_value:
                    ix_list.append(ix)
                    name_list.append(variable.name)
            else:
                logger.error(f'Invalid equality type ({equality_type}) passed to '
                             f'get_variables_where. Only values "equals" or "in" are valid.')
                raise ValueError

        if return_type == 'ix':
            return ix_list
        elif return_type == 'name':
            return name_list
        else:
            logger.error('Unknown value passed as return_type in get_variables_where.')
            raise ValueError

    def get_names(self):
        '''Get list of variable nicknames.

        Returns:
            list of variable nicknames
        '''

        variable_names = [variable.name for variable in self.varlist]

        return variable_names

    def get_columns_to_read(self, data_file):
        '''Get the name(s) of the column(s) from the input file that need to be read in order to
        create the all-names file.

        Args:
            data_file (DataFile object): contains info about a given input file

        Return:
            list of column names
        '''

        all_column_names = [data_file.record_id_col]
        for variable in self.varlist:
            column_names = variable.get_columns_to_read(data_file.nickname)
            all_column_names.extend([column_names] if type(column_names) == str else column_names)

        return list(set(all_column_names))

    def get_an_column_names(self):
        '''Get the final list of all-names columns, including internally created columns
        like `file_type` and `drop_from_nm`.

        Return:
            list of all-names columns
        '''

        variable_list = self.copy().varlist

        column_names = ["record_id", "file_type", "drop_from_nm", "dataset"]
        for variable in variable_list:
            if variable.compare_type == 'Address':
                column_names.extend(
                        ['address_street_number',
                        'address_street_name',
                        'address_street_type'])
            else:
                column_names.append(variable.name)

        return column_names

    def write(self, output_file):
        '''Write the VariableList to a yaml file.

        Args:
            output_file (str): path to write variable list dictionary
        '''

        dump_yaml(to_dict(self), output_file)

    def copy(self):
        '''Create a deep copy of a VariableList object.'''

        return deepcopy(self)



import tempfile

from contextlib import contextmanager

import pandas as pd


@contextmanager
def make_temp_parquet_file(file_path):
    df = pd.read_csv(file_path)
    tf = tempfile.NamedTemporaryFile(delete=False)
    df.to_parquet(tf.name)
    yield tf.name
    tf.close()




import csv
import logging
import numpy as np
import pandas as pd

import pyarrow as pa
import pyarrow.parquet as pq

from namematch.base import NamematchBase
from namematch.data_structures.schema import Schema
from namematch.data_structures.parameters import Parameters
from namematch.utils.utils import log_runtime_and_memory, load_yaml
from namematch.utils.profiler import Profiler

profile = Profiler()
logger = logging.getLogger()


class GenerateMustLinks(NamematchBase):
    def __init__(
        self,
        params,
        schema,
        all_names_file,
        must_links,
        *args,
        **kwargs
    ):

        super(GenerateMustLinks, self).__init__(params, schema, *args, **kwargs)
        self.must_links = must_links
        self.all_names_file = all_names_file

    @property
    def output_files(self):
        return [
            self.must_links
        ]

    # @log_runtime_and_memory
    def main(self, **kw):
        '''Generate the list of must-link pairs using UniqueID and ExistingID info .

        Args:
            params (Parameters object): contains parameter values
            schema (Schema object): contains match schema info (files to match, variables to use, etc.)
            all_names_file (str): path to the all-names file
            must_links (str): path to the must-links file
        '''
        logger.info('Generating "must-link" record pairs.')
        # get UniqueID variables
        uid_vars_list = self.schema.variables.get_variables_where(
                attr='compare_type', attr_value='UniqueID')

        # get ExistingID variables (incremental)
        eid_vars_list = self.schema.variables.get_variables_where(
                attr='compare_type', attr_value='ExistingID')

        # get records with non-missing unique identifiers
        ml_var_df = self.build_ml_var_df(
                self.all_names_file, uid_vars_list, eid_vars_list)

        # get the "must-link" record pairs
        must_links_df = self.get_must_links(
                ml_var_df, uid_vars_list, eid_vars_list)

        # true record pairs
        must_links_df.to_csv(
            self.must_links,
            index=False,
            quoting=csv.QUOTE_NONNUMERIC
        )

        if self.enable_lprof:
            self.write_line_profile_stats(profile.line_profiler)

    def build_ml_var_df(self, all_names_file, uid_vars_list, eid_vars_list, **kw):
        '''Load the all-names file and limit it to the rows that have either
        a non-missing UniqueID or ExistingID value.

        Args:
            all_names_file (str): path to the all-names file
            uid_vars_list (list of strings): all-name columns with compare_type "UniqueID"
            eid_vars_list (list of strings): all-name columns with compare_type "ExistingID"

        Returns:
            pd.DataFrame: a subset of the all-names file, relevant colums only

            ======================   =======================================================
            record_id                unique record identifier
            blockstring              concatenation of all the blocking variables (sep by ::)
            file_type                either "new" or "existing"
            drop_from_nm             flag, 1 if met any "to drop" criteria 0 otherwise
            <UniqueID column(s)>     variables of compare_type UniqueID
            <ExistingID column(s)>   variables of compare_type ExistingID
            has_ml_var               flag, always 1 in output (ml stands for must-link)
            ======================   =======================================================
        '''

        cols = ['record_id', 'blockstring', 'drop_from_nm',
                'file_type'] + uid_vars_list + eid_vars_list

        table = pq.read_table(all_names_file)
        an = table.to_pandas()[cols]

        an['has_ml_var'] = 0
        for ml_var in (uid_vars_list + eid_vars_list):
            an.loc[an[ml_var] != '', 'has_ml_var'] = 1

        return an[an.has_ml_var == 1]

    # @log_runtime_and_memory
    @profile
    def get_must_links(self, ml_var_df, uid_vars_list, eid_vars_list, **kw):
        '''Expand the list of records with must-link information to pairs of records
        that must be linked togehter in the final match.

        Args:
            ml_var_df (pd.DataFrame):
                ======================   =======================================================
                record_id                unique record identifier
                blockstring              concatenation of all the blocking variables (sep by ::)
                file_type                either "new" or "existing"
                drop_from_nm             flag, 1 if met any "to drop" criteria 0 otherwise
                <UniqueID column(s)>     variables of compare_type UniqueID
                <ExistingID column(s)>   variables of compare_type ExistingID
                has_ml_var               flag, always 1 in output (ml stands for must-link)
                ======================   =======================================================

            uid_vars_list (list of strings): all-name columns with compare_type "UniqueID"
            eid_vars_list (list of strings): all-name columns with compare_type "ExistingID"

        Returns:
            pd.DataFrame: list of must-link record pairs

            ===================   =======================================================
            record_id_1           unique identifier for the first record in the pair
            record_id_2           unique identifier for the second record in the pair
            blockstring_1         blockstring for the first record in the pair
            blockstring_2         blockstring for the second record in the pair
            drop_from_nm_1        flag, 1 if the first record in the pair was not eligible for matching
            drop_from_nm_2        flag, 1 if the second record in the pair was not eligible for matching
            existing              flag, 1 if the pair is must-link because of ExistingID
            ===================   =======================================================
        '''

        ml_var_df = ml_var_df.copy()

        # for each UniqueID variable, merge the data frame on itself to get must-links
        must_link_df_list = []
        for ml_var in (uid_vars_list + eid_vars_list):

            # warn if any uids are used more than n times (might be
            # sign of misspecified missing uid)
            if ml_var in uid_vars_list:
                uid_counts = ml_var_df[ml_var_df[ml_var] != ''].groupby(ml_var).size()
                uid_counts_high = uid_counts[uid_counts > 200].index.tolist()
                if len(uid_counts_high) > 0:
                    logger.warning(f"The following {ml_var} values have over 200 unique "
                                   f"values; please ensure strings such as 'NA' are not "
                                   f"getting coded as values.")
                    logger.info(uid_counts_high)
                if uid_counts.max() > 1000:
                    raise()

            must_link_df = pd.merge(
                    ml_var_df[ml_var_df[ml_var] != ''], ml_var_df,
                    on=ml_var, suffixes=['_1', '_2'])

            must_link_df = must_link_df[
                    (must_link_df.blockstring_1 < must_link_df.blockstring_2) |
                    ((must_link_df.blockstring_1 == must_link_df.blockstring_2) &
                     (must_link_df.record_id_1 < must_link_df.record_id_2))]

            must_link_df['existing'] = 0

            # tweak for incremental matching
            if len(eid_vars_list) > 0:
                if ml_var in uid_vars_list:
                    must_link_df = must_link_df[
                            (must_link_df.file_type_1 == 'new') |
                            (must_link_df.file_type_2 == 'new')]
                else:
                    must_link_df['existing'] = 1

            must_link_df = must_link_df[
                    ['record_id_1', 'record_id_2', 'blockstring_1', 'blockstring_2',
                    'drop_from_nm_1', 'drop_from_nm_2', 'existing']]

            must_link_df_list.append(must_link_df.copy())

        must_link_df = pd.concat(must_link_df_list)
        must_link_df = must_link_df.sort_values(['existing'], ascending=[False])
        # NOTE: existing == 1 takes priority over existing == 0

        # remove duplicates (would occur if records match on multiple Unique IDs)
        len_pre_drop_duplicates = len(must_link_df)
        must_link_df = must_link_df.drop_duplicates(subset=['record_id_1', 'record_id_2'])
        len_post_drop_duplicates = len(must_link_df)
        if (len_pre_drop_duplicates > len_post_drop_duplicates):
            logger.trace(f'Dropped {len_pre_drop_duplicates - len_post_drop_duplicates} '
                         f'duplicate ground truth rows')

        return must_link_df




__version__ = '1.2.1'
import os
import yaml

default_parameters = yaml.load(open(os.path.join(os.path.dirname(__file__), 'default_parameters.yaml'), 'r'), Loader=yaml.FullLoader)
default_parameters = dict((k,v) if v != 'None' else (k, None) for k,v in default_parameters.items())




import numpy as np
import os
from collections import defaultdict
import pickle
from .counter import _editCounts, _ngramCount, _probName, _condProbName, _probSamePerson
import importlib

class NameProbability():
    def __init__(self, name_list=None, ngram_len=5, smoothing=.001, unique=0,
        edit_count_max=None, last_comma_first=0, name_list_location=None,
        save_location=None, use_SS=0):
        '''
        - edit_count_max is used to limit the number of samples to consider
        when computing edit operation counts
        '''
        self.smoothing = smoothing
        self.ngram_count = defaultdict(int)
        self.edit_count = defaultdict(int)
        self.ngram_len = ngram_len
        self.pop_size = 0
        self.total_edits = 0
        self.unique = unique
        self.last_comma_first = last_comma_first
        self.edit_count_max = edit_count_max
        self.DATA_PATH = 1
        self.memoize = defaultdict(float)
        self.cp_memoize = defaultdict(float)
        self.psp_memoize = defaultdict(float)
        self.name_list_location = name_list_location
        self.save_location = save_location
        self.name_list = name_list

        if not name_list and not name_list_location and not use_SS:
            raise Exception('Need either a name list, location to name list, or to use SS data')
        if name_list and name_list_location:
            raise Exception('Only one of name_list or name_list_location can be provided')
        name_list_provided = True if name_list or name_list_location else False

        if name_list_provided:
            if name_list_location:
                self.loadNameList()

            if not isinstance(name_list, list):
                self.name_list = list(self.name_list)
            if self.unique:
                self.name_list = list(set(self.name_list))
            if self.last_comma_first:
                for i, n in enumerate(self.name_list):
                    rev_name = n.split(', ')
                    rev_name = rev_name[1] + ' ' + rev_name[0]
                    self.name_list[i] = rev_name

            self.pop_size += len(self.name_list)
            self.ngramCount(self.name_list)
            self.editCounts(self.name_list)

        if not name_list_provided and not use_SS:
            raise Exception('No training data provided. Use either a custom name list or the Social Security data')

        if not name_list_provided and use_SS:
            DATA_PATH = importlib.resources.read_binary(__file__, 'ss_daata.pkl')
            with open(DATA_PATH, 'r') as f:
                ss_data = pickle.load(f)
                self.ngram_count = ss_data[0]
                self.edit_count = ss_data[1]
                self.pop_size = 25e6
                self.total_edits = sum(v for v in self.edit_count.itervalues())


    def ngramCount(self, name_list):
        self.ngram_count = _ngramCount(name_list, self.ngram_len)


    def editCounts(self, name_list):
        # to compute probability of edit operations use a subsample of names
        if self.edit_count_max:
            name_list = np.array(name_list)
            name_samp = name_list[np.random.randint(0, len(name_list),
                                                    self.edit_count_max)].tolist()
        else:
            name_samp = name_list
        edit_count, total_edits = _editCounts(name_samp)
        self.total_edits += total_edits
        for k, v in edit_count.items():
            self.edit_count[k] += v


    def probName(self, name):
        # compute the probability of name based on the training data
        if len(name) < self.ngram_len:
            return 0
        if name not in self.memoize:
            self.memoize = _probName(name, self.ngram_len, self.ngram_count,
                                     self.smoothing, self.memoize)
        return self.memoize[name]


    def condProbName(self, name1, name2):
        # computes the conditional probability of arriving at name1
        # by performing a series of operation on name2.
        if (name1, name2) not in self.cp_memoize:
            self.cp_memoize = _condProbName(name1, name2, self.edit_count, self.total_edits,
                                            self.smoothing, self.cp_memoize)
        return self.cp_memoize[(name1, name2)]


    def probSamePerson(self, name1, name2):
        # computes the probability that the two names belong to the same person.
        if len(name1) < self.ngram_len or len(name2) < self.ngram_len:
            print('Both names should be at least', self.ngram_len, ' characters long')
            return 0.0
        if (name1, name2) not in self.psp_memoize:
            self.psp_memoize, self.cp_memoize, self.memoize = _probSamePerson(name1,
                           name2, self.pop_size, self.edit_count,
                           self.total_edits, self.smoothing, self.ngram_len,
                           self.ngram_count, self.memoize, self.cp_memoize,
                           self.psp_memoize)
        return self.psp_memoize[(name1, name2)]


    def probUnique(self, name):
        # compute the probability that a name is unique in the data
        return 1. / ((self.pop_size - 1) * self.probName(name) + 1)

    def surprisal(self, name):
        return -np.log2(self.probUnique(name))

    def saveObject(self):
        with open(self.save_location, 'w') as f:
            pickle.dump(temp, f)

    def loadNameList(self):
        with open(self.name_list_location, 'r') as f2:
            self.name_list = f2.read().split('\n')



import logging
import os
from pathlib import Path
import ruamel.yaml

from typing import Union
from shutil import copyfile

from namematch.data_structures.schema import Schema
from namematch.data_structures.parameters import Parameters
from namematch.utils.utils import to_dict, setup_logging, load_logging_params, reformat_dict
from namematch import default_parameters
from namematch.process_input_data import ProcessInputData
from namematch.generate_must_links import GenerateMustLinks
from namematch.block import Block
from namematch.generate_data_rows import GenerateDataRows
from namematch.fit_model import FitModel
from namematch.predict import Predict
from namematch.cluster import Cluster, Constraints
from namematch.generate_output import GenerateOutput
from namematch.generate_report import GenerateReport

yaml = ruamel.yaml.YAML()
logging.basicConfig(
    format="%(asctime)s %(levelname)-8s %(message)s",
    datefmt="%d-%b-%y %H:%M:%S",
    level=logging.INFO
)


class NameMatcher(object):
    """Main interface to run all the steps in namematch"""
    def __init__(
        self,
        # inputs
        config: dict=None,
        default_params: dict=None,
        og_blocking_index_file: str='None',
        trained_model_info_file: str='None',
        # what should outputs be called
        nm_info_file: str='nm_info.yaml',
        log_file_name: str=None,
        logging_params_file: str=None,
        output_dir: str='output',
        output_temp_dir: str=None,
        all_names_file: str='all_names.parquet',
        must_links: str='must_links.csv',
        blocking_index_bin_file: str='blocking_index.bin',
        candidate_pairs_file: str='candidate_pairs.parquet',
        data_rows_dir: str='data_rows',
        selection_model_name: str='basic_selection_model.pkl',
        match_model_name: str='basic_match_model.pkl',
        flipped0_file: str='flipped0_potential_links.csv',
        model_dir: str='model',
        model_info_file: str='model.yaml',
        potential_edges_dir: str='potential_links',
        cluster_assignments: str='cluster_assignments.pkl',
        edges_to_cluster: str='edges_to_cluster.parquet',
        constraints: Union[str, Constraints]=None,
        an_output_file: str='all_names_with_clusterid.csv',
        report_file: str='matching_report.html',
        # profilers
        enable_lprof: bool=False,
        logging_level: str='INFO',
        # for reload
        params=None,
        schema=None,
    ):

        self.config = config
        self.default_params = default_params if default_params else default_parameters

        # output
        self.output_dir = os.path.abspath(output_dir)

        if output_temp_dir:
            self.output_temp_dir = os.path.abspath(output_temp_dir)
        else:
            self.output_temp_dir = os.path.join(self.output_dir, 'details')

        # profilers
        self.enable_lprof = enable_lprof
        self.profile_dir = os.path.join(self.output_temp_dir, 'profilers')

        # logging files
        self.log_file_name = log_file_name if log_file_name else 'name_match.log'
        self.log_file = os.path.join(self.output_temp_dir, self.log_file_name)

        logging.info(f"The log file will be located at {self.log_file}.")
        self.logging_params_file = logging_params_file
        self.logging_params = load_logging_params(self.logging_params_file)

        # nm stats yaml file
        self.nm_info_file = nm_info_file
        self.nm_info_file_path = os.path.join(self.output_temp_dir, self.nm_info_file)

        # schema and params
        self.params = params
        self.schema = schema

        # incremental inputs
        self.og_blocking_index_file = og_blocking_index_file
        self.trained_model_info_file = trained_model_info_file

        # all the intermediates in output/details
        self._all_names_file = all_names_file
        self._must_links = must_links
        self._candidate_pairs_file = candidate_pairs_file
        self._data_rows_dir = data_rows_dir
        self._flipped0_file = flipped0_file
        self._selection_model_name = selection_model_name
        self._match_model_name = match_model_name
        self._model_dir = model_dir
        self._model_info_file = model_info_file
        self._potential_edges_dir = potential_edges_dir
        self._cluster_assignments = cluster_assignments
        self._edges_to_cluster = edges_to_cluster
        self._constraints = constraints
        self._an_output_file = an_output_file
        self._report_file = report_file

        # prepend output/details path to output files
        self.all_names_file = os.path.join(self.output_temp_dir, all_names_file)
        self.must_links = os.path.join(self.output_temp_dir, must_links)
        self.blocking_index_bin_file = os.path.join(self.output_temp_dir, blocking_index_bin_file)
        self.candidate_pairs_file = os.path.join(self.output_temp_dir, candidate_pairs_file)
        self.data_rows_dir = os.path.join(self.output_temp_dir, data_rows_dir)
        self.model_dir = os.path.join(self.output_temp_dir, self._model_dir)
        self.selection_model_name = selection_model_name
        self.match_model_name = match_model_name
        self.flipped0_file = os.path.join(self.output_temp_dir, self._flipped0_file)
        self.model_info_file = os.path.join(self.model_dir, self._model_info_file)
        self.potential_edges_dir = os.path.join(self.output_temp_dir, self._potential_edges_dir)
        self.cluster_assignments = os.path.join(self.output_temp_dir, self._cluster_assignments)
        self.edges_to_cluster = os.path.join(self.output_temp_dir, self._edges_to_cluster)
        self.constraints = constraints
        self.an_output_file = os.path.join(self.output_temp_dir, self._an_output_file)

        # prepend output path
        self.report_file = os.path.join(self.output_dir, self._report_file)

        # Create the folder structure
        if not os.path.exists(self.output_dir):
            os.makedirs(self.output_dir)

        if not os.path.exists(self.output_temp_dir):
            os.makedirs(self.output_temp_dir)

        if not os.path.exists(self.data_rows_dir):
            os.makedirs(self.data_rows_dir)

        if not os.path.exists(self.model_dir):
            os.makedirs(self.model_dir)

        if not os.path.exists(self.potential_edges_dir):
            os.makedirs(self.potential_edges_dir)

        if self.enable_lprof and (not os.path.exists(self.profile_dir)):
            os.makedirs(self.profile_dir)

        # process config
        if not self.params or not self.schema:
            self._process_config()
        else:
            logging.info("params and schema are provided. don't have to process config")

        # Create the nm_info file
        if not os.path.exists(self.nm_info_file_path):
            nm_metadata = self.nm_metadata
            nm_info = {}
            for key, value in nm_metadata.items():
                nm_metadata[key] = reformat_dict(value)

            nm_info['metadata'] = nm_metadata
            nm_info['stats'] = {}
            with open(self.nm_info_file_path, 'w') as f:
                yaml.dump(nm_info, f)

        # Set up logger
        setup_logging(
            self.logging_params,
            self.log_file,
            self.output_temp_dir,
            filter_stats=False,
            logging_level=logging_level,
        )
        self.logging_params['filters']['stat_filter']['()'] = 'StatLogFilter'


        # Instantiate each step
        self._process_input_data = ProcessInputData(
            params=self.params,
            schema=self.schema,
            all_names_file=self.all_names_file,
            nm_info_file=self.nm_info_file_path,
            nm_metadata=self.nm_metadata,
            profile_dir=self.profile_dir,
            enable_lprof=self.enable_lprof,
        )

        self._generate_must_links = GenerateMustLinks(
            params=self.params,
            schema=self.schema,
            all_names_file=self.all_names_file,
            must_links=self.must_links,
            nm_info_file=self.nm_info_file_path,
            nm_metadata=self.nm_metadata,
            profile_dir=self.profile_dir,
            enable_lprof=self.enable_lprof,
        )

        self._block = Block(
            params=self.params,
            schema=self.schema,
            all_names_file=self.all_names_file,
            must_links_file=self.must_links,
            og_blocking_index_file=self.og_blocking_index_file,
            blocking_index_bin_file=self.blocking_index_bin_file,
            candidate_pairs_file=self.candidate_pairs_file,
            nm_info_file=self.nm_info_file_path,
            nm_metadata=self.nm_metadata,
            profile_dir=self.profile_dir,
            enable_lprof=self.enable_lprof,
        )

        self._generate_data_rows = GenerateDataRows(
            params=self.params,
            schema=self.schema,
            all_names_file=self.all_names_file,
            candidate_pairs_file=self.candidate_pairs_file,
            output_dir=self.data_rows_dir,
            nm_info_file=self.nm_info_file_path,
            nm_metadata=self.nm_metadata,
            profile_dir=self.profile_dir,
            enable_lprof=self.enable_lprof,
        )

        self._fit_model = FitModel(
            params=self.params,
            all_names_file=self.all_names_file,
            data_rows_dir=self.data_rows_dir,
            trained_model_info_file=self.trained_model_info_file,
            selection_model_name=self.selection_model_name,
            match_model_name=self.match_model_name,
            flipped0_file=self.flipped0_file,
            model_info_file=self.model_info_file,
            output_dir=self.model_dir,
            nm_info_file=self.nm_info_file_path,
            nm_metadata=self.nm_metadata,
            profile_dir=self.profile_dir,
            enable_lprof=self.enable_lprof,
        )

        self._predict = Predict(
            params=self.params,
            data_rows_dir=self.data_rows_dir,
            model_info_file=self.model_info_file,
            output_dir=self.potential_edges_dir,
            nm_info_file=self.nm_info_file_path,
            nm_metadata=self.nm_metadata,
            profile_dir=self.profile_dir,
            enable_lprof=self.enable_lprof,
        )

        self._cluster = Cluster(
            params=self.params,
            schema=self.schema,
            constraints=self.constraints,
            must_links_file=self.must_links,
            potential_edges_dir=self.potential_edges_dir,
            flipped0_edges_file=self.flipped0_file,
            all_names_file=self.all_names_file,
            cluster_assignments=self.cluster_assignments,
            edges_to_cluster=self.edges_to_cluster,
            nm_info_file=self.nm_info_file_path,
            nm_metadata=self.nm_metadata,
            profile_dir=self.profile_dir,
            enable_lprof=self.enable_lprof,
        )

        self._generate_output = GenerateOutput(
            params=self.params,
            schema=self.schema,
            all_names_file=self.all_names_file,
            cluster_assignments_file=self.cluster_assignments,
            an_output_file=self.an_output_file,
            output_dir=self.output_dir,
            nm_info_file=self.nm_info_file_path,
            nm_metadata=self.nm_metadata,
            profile_dir=self.profile_dir,
            enable_lprof=self.enable_lprof,
        )

        self._generate_report = GenerateReport(
            params=self.params,
            schema=self.schema,
            report_file=self.report_file,
            nm_info_file=self.nm_info_file_path,
            nm_metadata=self.nm_metadata,
            profile_dir=self.profile_dir,
            enable_lprof=self.enable_lprof,
        )

        self._process_input_data.next = self._generate_must_links
        self._generate_must_links.next = self._block
        self._block.next = self._generate_data_rows
        self._generate_data_rows.next = self._fit_model
        self._fit_model.next = self._predict
        self._predict.next = self._cluster
        self._cluster.next = self._generate_output
        self._generate_output.next = self._generate_report

    @property
    def process_input_data(self):
        self._process_input_data.nm_metadata = self.nm_metadata

        return self._process_input_data

    @property
    def generate_must_links(self):
        self._generate_must_links.nm_metadata = self.nm_metadata
        return self._generate_must_links

    @property
    def block(self):
        self._block.nm_metadata = self.nm_metadata
        return self._block

    @property
    def generate_data_rows(self):
        self._generate_data_rows.nm_metadata = self.nm_metadata
        return self._generate_data_rows

    @property
    def fit_model(self):
        self._fit_model.nm_metadata = self.nm_metadata
        return self._fit_model

    @property
    def predict(self):
        self._predict.nm_metadata = self.nm_metadata
        return self._predict

    @property
    def cluster(self):
        self._cluster.nm_metadata = self.nm_metadata
        return self._cluster

    @property
    def generate_output(self):
        self._generate_output.nm_metadata = self.nm_metadata
        return self._generate_output

    @property
    def generate_report(self):
        self._generate_report.nm_metadata = self.nm_metadata
        return self._generate_report

    @property
    def all_tasks(self):
        tasks = [
            "process_input_data",
            "generate_must_links",
            "block",
            "generate_data_rows",
            "fit_model",
            "predict",
            "cluster",
            "generate_output",
            "generate_report",
        ]
        for task in tasks:
            yield getattr(self, task)

    @property
    def nm_metadata(self):
        '''Namematch state including all the necessary attributes to recreate the NameMatcher object'''
        metadata = {}
        # metadata['metadata'] = {}

        params = to_dict(self.params)
        schema = to_dict(self.schema)

        metadata['params'] = params
        metadata['schema'] = schema
        metadata['filepaths'] = {}
        for key, value in self.__dict__.items():
            if isinstance(value, (str, dict, bool, int, float)) or value is None:
                if key not in params or key not in schema:
                    if key not in ["config", "default_params"]:
                        metadata['filepaths'][key] = value

        return metadata

    def _process_config(self):
        """Process config file to create parameter and schema"""
        self.params = Parameters.init(self.config, self.default_params)
        self.schema = Schema.init(self.config, self.params)
        self.params.validate(self.schema.variables)

    def run(self, force=False, write_params_schema_file=True, write_stats_file=True):
        """Main method to kick off the namematch process

        Args:
            force (bool): Force to run all the tasks
            write_params_schema_file (bool): whether to write params and schema to yaml
            write_stats_file (bool): whether to write the nm_info file
        """
        for task in self.all_tasks:
            task.run(force=force, write_stats_file=write_stats_file)

    @classmethod
    def load_namematcher(cls, nm_info_file_path, new_nm_info_file=None, **kwargs):
        '''To load NameMatcher instance given the nm_info_file.
        This classmethod help us pick up where it left last time based on the information in the
        nm_info_file. It will create a NameMatcher instance and recover all the attributes as
        well as stats_dict for tasks that was already run.

        Args:
            nm_info_file (str): nm_info.yaml file path

        Returns:
            :mod:`namematch.namematcher.NameMatcher`: NameMatcher instance

        '''
        with open(nm_info_file_path) as f:
            nm_info = yaml.load(f)
            stats = nm_info['stats']
            metadata = nm_info['metadata']
            filepaths = metadata['filepaths']

        params = Parameters.load_from_dict(metadata['params'])
        schema = Schema.load_from_dict(metadata['schema'])

        if new_nm_info_file:
            copyfile(nm_info_file_path, os.path.join(os.path.dirname(nm_info_file_path), new_nm_info_file))

        for key, value in kwargs.items():
            if key in filepaths:
                filepaths[key] = value

        nm = cls(
            nm_info_file=new_nm_info_file if new_nm_info_file else Path(nm_info_file_path).name,
            log_file_name=filepaths['log_file_name'],
            logging_params_file=None,
            output_dir=filepaths['output_dir'],
            output_temp_dir=filepaths['output_temp_dir'],
            all_names_file=filepaths['_all_names_file'],
            must_links=filepaths['_must_links'],
            og_blocking_index_file=filepaths['og_blocking_index_file'],
            candidate_pairs_file=filepaths['_candidate_pairs_file'],
            data_rows_dir=filepaths['_data_rows_dir'],
            trained_model_info_file=filepaths['trained_model_info_file'],
            selection_model_name=filepaths['selection_model_name'],
            match_model_name=filepaths['match_model_name'],
            flipped0_file=filepaths['_flipped0_file'],
            model_dir=filepaths['_model_dir'],
            model_info_file=filepaths['_model_info_file'],
            potential_edges_dir=filepaths['_potential_edges_dir'],
            cluster_assignments=filepaths['_cluster_assignments'],
            constraints=filepaths['_constraints'],
            an_output_file=filepaths['_an_output_file'],
            report_file=filepaths['_report_file'],
            params=params,
            schema=schema,
            logging_level='DEBUG',
        )
        for task in nm.all_tasks:
            task_name = task.__class__.__name__
            if task_name in stats:
                task.stats_dict = stats[task_name]

        return nm



import pytest
import tempfile
import logging
import pickle

from contextlib import contextmanager

import pandas as pd

from namematch.data_structures.schema import Schema
from namematch.data_structures.parameters import Parameters
from namematch.utils.utils import setup_logging

from tests.utils import make_temp_parquet_file

PATH = "tests/unit/data/"

@pytest.fixture
def default_params_dict():
    return {
        # basic
        'verbose' : None,
        'num_workers' : 1,
        'parallelize' : False,
        # read-in rate (memory tradeoff)
        'input_data_batch_size': 50000,
        'data_rows_batch_size': 500,
        # variables
        'required_variables': ['first_name', 'last_name', 'dob'],
        'first_name_column': 'first_name',
        'last_name_column': 'last_name',
        'exact_match_variables': ['first_name', 'last_name', 'dob'],
        'negate_exact_match_variables' : ['middle_initial'],
        # data
        'split_names': True,
        'auto_drop_logic': [{'first_name': 'JOHN', 'last_name': 'DOE'},
                            {'first_name': 'JANE', 'last_name': 'DOE'}],
        # blocking
        'blocking_scheme': {'cosine_distance': {'variables': ['first_name','last_name']},
        'edit_distance': {'variable': 'dob'},
        'absvalue_distance': {'variable': 'age'},
        'alpha': 1.4,
        'power': 0.1},
        'blocking_thresholds': {
            'common_name_max_penalty': 0.1,
            'nodob_cosine_bar': 0.26,
            'high_cosine_bar': 0.3,
            'low_cosine_bar': 0.4,
            'high_editdist_bar': 1,
            'low_editdist_bar': 2,
            'absvalue_bar': 3
        },
        'index': {
            'rebuild_main_index': 'if_secondary_index_exceeds_limit',
            'secondary_index_limit': 500000
        },
        'nmslib': {'M': 100, 'efC': 1000, 'post': 0, 'efS': 750, 'k': 500},
        # modeling
        'pct_train' : 0.9,
        'use_uncovered_phats': False,
        'missingness_model': 'dob',
        'max_match_train_n': 3000000,
        'max_selection_train_eval_n': 1000000,
        'weight_using_selection_model': False,
        'default_threshold': 0.7,
        'optimize_threshold': True,
        'match_train_criteria': {'data_rows': {'covered_pair': 1}},
        'initialize_from_ground_truth_1s': True,
        'fscore_beta': 0.5,
        'allow_clusters_w_multiple_unique_ids': False
    }


@pytest.fixture
def config_dict():
    return {
        # input data files
        'data_files': {
            'raw_data': {
                'filepath': 'tests/unit/data/raw_data.csv',
                'record_id_col': 'row_id',
                'delim': ',',
                'cluster_type': 'cluster',
                'output_name': 'input_data'
            }
        },
        'variables': [
            {
                'name': 'first_name',
                'compare_type': 'String',
                'raw_data_col': 'firstname',
                'drop': ['', ' ', 'FIRM', 'BUSINESS', 'STATE OF', 'UNIVERSITY']
            },
            {
                'name': 'last_name',
                'compare_type': 'String',
                'raw_data_col': 'lastname'
            },
            {
                'name': 'dob',
                'compare_type': 'Date',
                'raw_data_col': 'birthdate',
                'check': 'Date - %m/%d/%Y'
            },
            {
                'name' : 'age',
                'compare_type': 'Numeric',
                'raw_data_col' : 'age_2021',
            },
            {
                'name': 'gender',
                'compare_type': 'Categorical',
                'raw_data_col': 'gender',
                'check': 'm,f'
            },
            {
                'name': 'address',
                'compare_type': 'Address',
                'raw_data_col': 'address'
            },
            {
                'name': 'uid',
                'compare_type': 'UniqueID',
                'raw_data_col': 'uid'}
        ],
        'verbose': 50000,
        'num_workers': 10,
        'allow_clusters_w_multiple_unique_ids': False,
        'leven_thresh': None,
        'pct_train': 0.6,
        'missingness_model': None
    }

@pytest.fixture
def logger_for_testing():
    logging_params_dict = {
        'version': 1,
        'disable_existing_loggers': False,
        'formatters': {
            'simple': {'format': '%(levelname)-8s %(message)s'},
            'colored_console': {
                '()': 'coloredlogs.ColoredFormatter',
                'format': "%(asctime)s - %(levelname)-8s %(message)s",
            },
            'detailed': {
                'format': '%(asctime)s %(levelname)-8s %(message)s',
                'datefmt': '%m/%d/%Y %H:%M:%S'
            },
            'message': {'format': '%(message)s'}
        },
        'filters': {
            'stat_filter': {'()': 'StatLogFilter'}
        },
        'handlers': {
            'console': {
                'class': 'logging.StreamHandler',
                'level': 'INFO',
                'formatter': 'colored_console',
            'stream': 'ext://sys.stdout'
            },
            'file_handler': {
                'class': 'logging.handlers.WatchedFileHandler',
                'level': 'DEBUG',
                'formatter': 'detailed',
                'encoding': 'utf8'
            },
            'file_handler_stat_memory': {
                'class': 'logging.StreamHandler',
                'level': 'STAT',
                'formatter': 'message'
            }
        },
        'root': {
            'level': 'DEBUG',
            'handlers': ['console', 'file_handler']
        }
    }

    log_file = tempfile.NamedTemporaryFile(mode='w+', suffix='.log', delete=False)
    setup_logging(logging_params_dict, log_file.name, tempfile.gettempdir())
    logging_params_dict['filters']['stat_filter']['()'] = 'StatLogFilter'
    logger = logging.getLogger()
    logging.disable(logging.CRITICAL)
    return logger


@pytest.fixture
def raw_data_df():
    df = pd.read_csv(PATH + "raw_data.csv")
    return df


@pytest.fixture
def an_df():
    df = pd.read_csv(PATH + "an.csv")
    return df


@pytest.fixture
def params_and_schema(config_dict, default_params_dict):
    # test passing
    params = Parameters.init(
        config_dict,
        default_params_dict)

    schema = Schema.init(config_dict, params)

    return params, schema


@pytest.fixture
def all_names_parquet_file():
    with make_temp_parquet_file(PATH + "an.csv") as an_fn:
        yield an_fn


@pytest.fixture
def blocking_exact_df():
    df_exact = pd.read_csv(PATH + "apply_blocking_exact.csv")
    return df_exact


@pytest.fixture
def blocking_notexact_df():
    df_notexact = pd.read_csv(PATH + "apply_blocking_notexact.csv")
    return df_notexact


@pytest.fixture
def nn_string_full_df():
    full_info = pd.read_csv(PATH + "nn_string_full_df.csv").set_index("nn_string")
    return full_info


@pytest.fixture
def nn_string_ed_string_df():
    ed_info = pd.read_csv(PATH + "nn_string_ed_string_df.csv").set_index("nn_string")
    return ed_info


@pytest.fixture
def thresholds_dict():
    return {
        "high_editdist_bar": 5,
        "low_editdist_bar": 0,
        "high_cosine_bar": 0.5,
        "nodob_cosine_bar": 0.5,
        "low_cosine_bar": 0,
        "absvalue_bar": 3,
        "common_name_max_penalty": 0.1
    }


@pytest.fixture
def near_neighbors_df():
    near_neighbors_df = pd.read_csv(PATH + "near_neighbors_df.csv")
    return near_neighbors_df


@pytest.fixture
def nn_strings_to_query():
    nn_strings_to_query = pd.read_csv(PATH + "nn_string_info.csv").nn_string.tolist()
    return nn_strings_to_query


@pytest.fixture
def data_rows_parquet_file():
    with make_temp_parquet_file(PATH + "data_rows.csv") as data_rows_fn:
        yield data_rows_fn


@pytest.fixture
def all_names_clustering_df():
    an = pd.read_csv(PATH + "all_names_clustering.csv")
    return an


@pytest.fixture
def must_links_df():
    ml = pd.read_csv(PATH + "gt_edges_clustering.csv")
    return ml


@pytest.fixture
def side_by_side_df():
    df = pd.read_csv(PATH + "side_by_side_df.csv")
    return df


@pytest.fixture
def side_by_side_df_with_probs():
    df = pd.read_csv(PATH + "side_by_side_df_with_probs.csv")
    return df


@pytest.fixture
def ids_df():
    df = pd.read_csv(PATH + "ids.csv",dtype=object, na_filter=False)
    return df



import numpy as np
import pandas as pd


def is_valid_link(predicted_links_df):
    '''Check if two records would form a valid link. For this default version, 
    it simply returns True.

    Args: 
        predicted_links_df (pd.DataFrame): info on predicted links (record pairs)
            ================      =======================================================
            record_id_1           unique record identifier (for first in pair)
            record_id_2           unique record identifier (for second in pair)
            phat                  predicted probability of a record pair being a match
            original_order        original ordering 1-N (useful so gt is always on top of phat=1 cases)
            <other_cols>          columns from all-names that are required for constraint checking (will have _1 and _2 versions)
            ================      =======================================================

    Returns: 
        boolean or boolean pd.Series: True for default version
    '''

    return True


def is_valid_cluster(cluster, phat):
    '''Check if a proposed cluster is valid. For this default version,
    it simply returns True. The information you'll have access to for the
    cluster's records is determined by the column(s) specified in `get_columns_used`.

    Args:
        cluster (pd.DataFrame): all-names file (relevant columns only) records for the proposed cluster
        phat (float): phat value of the proposed link

    Returns:
        bool: True for default version
    '''

    return True


def apply_link_priority(valid_links_df):
    '''Adjust the order in which valid link will be considered for clustering. For the vast majority of 
    runs, the default behavior -- sorting by descending phat, or P(match) -- is best. 

    Args:
        valid_links_df (pd.DataFrame): info on valid predicted links (record pairs)
            ================      =======================================================
            record_id_1           unique record identifier (for first in pair)
            record_id_2           unique record identifier (for second in pair)
            phat                  predicted probability of a record pair being a match
            original_order        original ordering 1-N (useful so gt is always on top of phat=1 cases)
            <other_cols>          columns from all-names that are required for constraint checking (will have _1 and _2 versions)
            ================      =======================================================

    Returns:
        pd.DataFrame: same as input, with phat columns potentially adjusted
    '''

    valid_links_df = valid_links_df.sort_values(by=['phat', 'original_order'], ascending=[False, True])

    return valid_links_df


def get_columns_used():
    '''The `get_columns_used()` function returns either the string "all" (default) or a 
    dictionary mapping the names of data fields needed for constraint checking to the data 
    type they should be read in as. This can be useful to limit memory consumption during the 
    clustering step (by only reading in the data fields only needed in cluster logic). 

    Special note: if you want something to be treated as a date, use
    dtype "date" (will be converted to a valid pandas dtype internally).

    Returns:
        dict: empty for default version
    '''

    return "all"



import json
from unittest.mock import patch

import pandas as pd

from namematch.generate_data_rows import GenerateDataRows


def test_generate_name_probabilities_object(params_and_schema, an_df, logger_for_testing):
    params, schema = params_and_schema
    with patch('namematch.generate_data_rows.logger', logger_for_testing) as mock_debug:
        generate_data_rows = GenerateDataRows(
            params,
            schema,
            all_names_file=an_df,
            candidate_pairs_file=None,
            output_dir=None,
        )

        prob_obj = generate_data_rows.generate_name_probabilities_object(
            an=an_df,
            fn_col=None,
            ln_col="column_name"
        )
        assert prob_obj is None

        prob_obj = generate_data_rows.generate_name_probabilities_object(
            an=an_df,
            fn_col='first_name',
            ln_col='last_name',
        )
        assert prob_obj is not None


def test_find_valid_training_records():
    pass


def test_generate_actual_data_rows():
    pass


def test_generate_data_row_files():
    pass



import pandas as pd
import numpy as np
import pickle
import pytest

from collections import Counter
from unittest.mock import patch

from namematch.block import (
    Block,
    get_common_name_penalties,
    read_an,
    get_second_index_nn_strings,
)

from namematch.data_structures.parameters import Parameters


def test_get_params_and_schema(params_and_schema):
    params, schema = params_and_schema

    print(params)
    print(schema)
    assert params.verbose == 50000

def test_get_blocking_columns():

    #get_blocking_columns(blocking_scheme)
    pass


def test_read_an():

    #read_an(an_file, nn_cols, ed_col, absval_col)
    pass


def test_get_nn_string_counts():

    #get_nn_string_counts(an)
    pass


def test_get_common_name_penalties(an_df, logger_for_testing):

    # load fake data
    last_names = an_df.last_name
    max_penalty = 0.4
    bins = 3

    # get the threshholds
    with patch('namematch.block.logger', logger_for_testing) as mock_debug:
        common = get_common_name_penalties(last_names, max_penalty, bins)

    # test
    assert last_names.nunique() == len(common)
    assert bins == len(Counter(common.values()))
    for k in common.keys():
        assert common[k] <= max_penalty


def test_read_all_names_parquet(all_names_parquet_file):
    an = read_an(all_names_parquet_file, ['first_name', 'last_name'], 'dob', None)
    assert an.shape == (23, 9)


def test_split_last_names(params_and_schema, all_names_parquet_file, logger_for_testing):
    params, schema = params_and_schema
    an = read_an(all_names_parquet_file, ['first_name', 'last_name'], 'dob', None)

    with patch('namematch.process_input_data.logger', logger_for_testing) as mock_debug:
        block = Block(
            params,
            schema,
            all_names_file=all_names_parquet_file,
            must_links_file=None,
            og_blocking_index_file=None,
            candidate_pairs_file=None,
        )
        an_split_ln = block.split_last_names(an, 'last_name', params.blocking_scheme)

        assert "orig_last_name" in an_split_ln
        assert "orig_record" in an_split_ln


def test_convert_all_names_to_blockstring_info(params_and_schema, all_names_parquet_file, logger_for_testing):
    _, schema = params_and_schema
    an = read_an(all_names_parquet_file, ['first_name', 'last_name'], 'dob', None)

    with patch('namematch.block.logger', logger_for_testing) as mock_debug:
        params = Parameters({
                "last_name_column": 'last_name',
                "blocking_scheme": {
                    'cosine_distance' : { 'variables' : ['first_name', 'last_name'] },
                    'edit_distance' : { 'variable' : 'dob' }
                },
                'blocking_thresholds': {'common_name_max_penalty' : .1},
                'split_names' : True})

        block = Block(
            params,
            schema,
            all_names_file=all_names_parquet_file,
            must_links_file=None,
            og_blocking_index_file=None,
            candidate_pairs_file=None,
        )

        nn_string_info, nn_string_expanded_df = block.convert_all_names_to_blockstring_info(an, None, params)

        assert nn_string_info[nn_string_info.nn_string == 'BEN::SMITH'].n_new.iloc[0] == 2
        assert len(nn_string_info) == 19
        assert len(nn_string_expanded_df) == 21


def test_get_all_shingles():

    #get_all_shingles()
    pass


def test_generate_shingles_matrix(params_and_schema, all_names_parquet_file, logger_for_testing):
    params, schema = params_and_schema

    with patch('namematch.block.logger', logger_for_testing) as mock_debug:
        block = Block(
            params,
            schema,
            all_names_file=all_names_parquet_file,
            must_links_file=None,
            og_blocking_index_file=None,
            candidate_pairs_file=None,
        )

        nn_strings = ['JOHN::SMITH', 'TAYLOR::JOHNSON', 'JAMES::JOHNS']
        alpha = 0.5
        power = 1
        matrix_type = "test"
        verbose = False
        m = block.generate_shingles_matrix(nn_strings, alpha, power, matrix_type, verbose)

        # sum of m's rows should equal 1 + alpha (but also floating point error)
        summed = m.sum()
        assert pytest.approx(summed) == (1 + alpha) * len(nn_strings)


def test_prep_index():

    #prep_index()
    pass


def test_load_main_index_nn_strings():

    #load_main_index_nn_strings(og_blocking_index_file)
    pass


def test_load_main_index():

    #load_main_index(index_file)
    pass


def test_generate_index():

    #generate_index(nn_strings, num_workers, M, efC, post, alpha, power, print_progress=True)
    pass


def test_get_second_index_nn_strings():

    all_strings = ["a", "b", "c"]
    main_strings = ["c"]
    expected_strings = ["a", "b"]
    strings = get_second_index_nn_strings(all_strings, main_strings)
    assert sorted(expected_strings) == sorted(strings)


def test_save_main_index():

    #save_main_index(main_index, main_index_nn_strings, main_index_file)
    pass


def test_get_indices():

    #get_indices(params, all_nn_strings, og_blocking_index_file)
    pass


def test_apply_blocking_filter_exact(params_and_schema, blocking_exact_df, nn_string_full_df, nn_string_ed_string_df, thresholds_dict):

    expanded_info = nn_string_full_df.join(nn_string_ed_string_df) # TEMP
    # exact matches
    params, schema = params_and_schema
    out = Block(params, schema).apply_blocking_filter(blocking_exact_df, thresholds_dict, expanded_info, nns_match=True)
    expected_names = ['blockstring_1', 'blockstring_2', 'cos_dist', 'edit_dist', 'covered_pair']
    assert expected_names == list(out)
    assert out.shape[0] == out.covered_pair.sum()


def test_apply_blocking_filter_not_exact(params_and_schema, blocking_notexact_df, nn_string_full_df, nn_string_ed_string_df, thresholds_dict):
    # not exact matches-- could be tested more exhaustively
    expanded_info = nn_string_full_df.join(nn_string_ed_string_df)
    expected_names = ['blockstring_1', 'blockstring_2', 'cos_dist', 'edit_dist', 'covered_pair']
    params, schema = params_and_schema
    out = Block(params, schema).apply_blocking_filter(blocking_notexact_df, thresholds_dict, expanded_info, nns_match=False)
    assert expected_names == list(out)
    assert blocking_notexact_df.shape[0] >= out.shape[0]


def test_disallow_switched_pairs():

    #disallow_switched_pairs(df, incremental, nn_strings_to_query)
    pass


def test_get_actual_candidates(params_and_schema, near_neighbors_df, nn_string_full_df, nn_string_ed_string_df, nn_strings_to_query, thresholds_dict):
    expanded_info = nn_string_full_df.join(nn_string_ed_string_df) # TEMP
    start_ix_worker = 0
    end_ix_worker = near_neighbors_df.shape[0]
    incremental = False

    # get pairs
    params, schema = params_and_schema
    cand_pairs = Block(params, schema).get_actual_candidates(
        near_neighbors_df.iloc[start_ix_worker : end_ix_worker],
        expanded_info,
        nn_strings_to_query,
        thresholds_dict,
        incremental)

    # test
    assert cand_pairs.shape[0] <= near_neighbors_df.shape[0]


def test_get_near_neighbors_df(params_and_schema):
    # changing this test involves re-calculating the near_neighbors_list using
    # functions like generate_indices and get_all_shingles and knnQueryBatch
    near_neighbors_list = [
        (np.array([3, 4]), np.array([0., 0.18983728])),
        (np.array([4, 3]), np.array([1.1920929e-07, 1.8983728e-01])),
        (np.array([5, 6]), np.array([5.9604645e-08, 2.2517276e-01])),
        (np.array([6, 5]), np.array([0., 0.22517282])),
    ]

    k = len(near_neighbors_list[0][0])

    nn_string_info = pd.DataFrame({
        "nn_string": ["MALI::H", "ZUB::JELVEH", "J::SMITS", "JOHN::SMITH", "JON::SMITH", "ABBY::LI", "ZABBIE::LI"],
        "commonness_penalty": [0, 0, 0, 0, 0, 0, 0],
        "n_new": [3, 6, 1, 2, 6, 8, 9]
    })

    nn_strings_this_index = nn_string_info.nn_string.tolist()

    start_ix_batch = 3
    end_ix_batch = 8
    nn_strings_queried_this_batch = nn_strings_this_index[start_ix_batch:end_ix_batch]

    # get the df
    params, schema = params_and_schema
    near_neighbors_df = Block(params, schema).get_near_neighbors_df(
        near_neighbors_list,
        nn_string_info,
        nn_strings_this_index,
        nn_strings_queried_this_batch)

    # test
    assert k * len(nn_strings_queried_this_batch) == near_neighbors_df.shape[0]

    js = {"JOHN::SMITH", "JON::SMITH"}
    azl = {"ABBY::LI", "ZABBIE::LI"}
    for (i, row) in near_neighbors_df.iterrows():
        if (row.nn_string_1 == row.nn_string_2):
            assert 0 == pytest.approx(row.cos_dist, abs=1e-5)
        else:
            if (row.nn_string_1 in js):
                assert row.nn_string_2 in js
            elif (row.nn_string_2 in azl):
                assert row.nn_string_2 in azl


def test_get_exact_match_candidate_pairs():

    #get_exact_match_candidate_pairs(
    #    nn_string_info_multi, nn_string_expanded_df, blocking_thresholds)
    pass

def test_get_query_strings():

    #get_query_strings(nn_string_info, blocking_scheme)
    pass

def test_write_some_cps():

    #write_some_cps(cand_pairs, output_file, header=False)
    pass

def test_generate_candidate_pairs():

    # generate_candidate_pairs(
    #     params, nn_strings_to_query, shingles_to_query,
    #     nn_string_info, nn_string_expanded_df,
    #     main_index, main_index_nn_strings,
    #     second_index, second_index_nn_strings,
    #     output_file)
    pass

def test_generate_true_pairs():

    #generate_true_pairs(must_links_df)
    pass

def test_compute_cosine_sim():

    #compute_cosine_sim(blockstrings_in_pairs, pairs_df, shingles_matrix)
    pass

def test_evaluate_blocking():

    #evaluate_blocking(cp_df, tp_df, blocking_scheme)
    pass

def test_add_uncovered_pairs():

    #add_uncovered_pairs(candidate_pairs_df, uncovered_pairs_df)
    pass

def test_pandas_default_min():

    # point is to test that min() default skipna=True doesn't change
    # for some reason, explicitely passing skinpna=True makes
    # the function a lot slower

    df = pd.DataFrame(data={
        'a':[1, 1, 2, 3, 4],
        'b':[np.NaN, 3, 1, 5, np.NaN]
    })
    df = df.groupby('a').b.min()

    assert df.isnull().sum(), 1




import logging
import multiprocessing as mp
import numpy as np
import os
import pandas as pd

import pyarrow.parquet as pq
import pyarrow as pa

from namematch.base import NamematchBase
from namematch.utils.utils import setup_logging
from namematch.data_structures.parameters import Parameters
from namematch.utils.utils import load_models, log_runtime_and_memory, determine_model_to_use, load_yaml
from namematch.utils.profiler import Profiler

profile = Profiler()
logger = logging.getLogger()


# globals:
MATCH_COL = 1


class Predict(NamematchBase):
    '''
     Args:
        params (Parameters object): contains parameter values
        model_info_file (str): path to the data-rows dir
        data_rows_dir (str): path to the model info yaml file for a trained model
        output_dir (str): path to the potential-links dir

    '''
    def __init__(
        self,
        params,
        data_rows_dir,
        model_info_file,
        output_dir,
        *args,
        **kwargs
    ):
        super(Predict, self).__init__(params, None, *args, **kwargs)

        self.data_rows_dir = data_rows_dir
        self.model_info_file = model_info_file
        self.output_dir = output_dir

    @property
    def output_files(self):
        output_files = []
        for dr_file in self.dr_file_list:
            thread = dr_file.split('/')[-1].replace('data_rows_', '').replace('.parquet', '')
            output_files.append(os.path.join(self.output_dir, f'potential_links_{thread}.parquet'))
        return output_files

    @property
    def dr_file_list(self):
        return [
            os.path.join(self.data_rows_dir, dr_file) for dr_file in os.listdir(self.data_rows_dir)
        ]

    # @log_runtime_and_memory
    def main(self, **kw):
        '''Read in data-rows and predict (in parallel) for each unlabeled pair. Output
        the pairs above the threshold.
        '''

        if not os.path.exists(self.output_dir):
            os.mkdir(self.output_dir)

        # remove previous potential_edges files if they exists
        for potential_edges_file in os.listdir(self.output_dir):
            os.remove(os.path.join(self.output_dir, potential_edges_file))

        match_models, model_info = load_models(self.model_info_file)

        self.get_potential_edges_in_parallel(
                match_models,
                model_info,
                self.output_dir,
                self.params)

        if self.enable_lprof:
            self.write_line_profile_stats(profile.line_profiler)

    # @log_runtime_and_memory
    @profile
    def get_potential_edges(self, dr_file, match_models, model_info,
                output_dir, params, **kw):
        '''Read in data rows in chunks and predict as needed. Write (append)
        the edges above the threshold to the appropriate file.

        Args:
            dr_file (string): path to data file to predict for
            match_models (dict): maps model name (e.g. basic or no-dob) to a fit match model object
            model_info (dict): contains information about threshold
            output_dir (str): directory to place potential links
            params (Parameters obj):  contains parameter values (i.e. use_uncovered_phats)
        '''
        try:
            import pathlib
            p = pathlib.Path(dr_file)
            thread = p.parts[-1].replace('data_rows_', '').replace('.parquet', '')
            output_file = os.path.join(output_dir, f'potential_links_{thread}.parquet')
            
            pf = pq.ParquetFile(p)
            batch_size = params.predict_batch_size
            logger.debug(f"batch size: {batch_size}")
            for i, pq_df in enumerate(pf.iter_batches(batch_size=batch_size)):
                df = pq_df.to_pandas()
                df = df[df.labeled_data == 0]

                if len(df) == 0:
                    continue

                df['model_to_use'] = determine_model_to_use(df, model_info)

                df = self.predict(match_models, df, 'match')

                df['potential_edge'] = 0
                for model_name in df.model_to_use.unique():
                    threshold = model_info[model_name]['match_thresh']
                    df.loc[(df.model_to_use == model_name) &
                           (df[f'{model_name}_match_phat'] >= threshold), 'potential_edge'] = 1

                if not params.use_uncovered_phats:
                    df.loc[df.covered_pair == 0, 'potential_edge'] = 0

                table = pa.Table.from_pandas(df)
                if i == 0:
                    pqwriter = pq.ParquetWriter(output_file, table.schema)
                    parquet_schema = table.schema
                if not df.empty:
                    pqwriter.write_table(table)
            if pqwriter:
                pqwriter.close()
        except Exception as e:
            os.remove(p)
            logger.error(f"{p} failed.")
            raise e
            

    def get_potential_edges_in_parallel(self, match_models, model_info, output_dir, params):
        '''Dispatch the worker threads that will predict for unlabeled pairs in paralle.

        Args:
            match_models (dict): maps model name (e.g. basic or no-dob) to a fit match model object
            model_info (dict): dict with information about how to fit the model
            output_dir
            params (Parameters object): contains parameter values
        '''
        
        if params.parallelize:

            jobs = [
                mp.Process(
                    target = self.get_potential_edges,
                    args = (
                        dr_file,
                        match_models,
                        model_info,
                        output_dir,
                        params)) for dr_file in self.dr_file_list]
            for job in jobs:
                job.start()
            for job in jobs:
                t = job.join()
            failure_occurred = sum([job.exitcode != 0 for job in jobs])
            if failure_occurred:
                logger.error("Error occurred in %s worker(s)." % failure_occurred)
                raise Exception("Error occurred in %s worker(s)." % failure_occurred)

        else:

            for dr_file in self.dr_file_list:
                self.get_potential_edges(dr_file, match_models, model_info,
                        output_dir, params)

    @classmethod
    def predict(cls, models, df, model_type, oob=False, all_cols=False, all_models=True, prob_match_train=None):
        '''Use the trainined models to predict for pairs of records.

        Args:
            models (dict): maps model name (e.g. basic or no-dob) to a fit match model object
            df (pd.DataFrame): portion of the data-rows table, with a "model_to_use" column appended
            model_type (str): model type (e.g. selection or match)
            oob (bool): if True, use the out-of-bag predictions
            all_cols (bool): if True, keep all columns in the output df; not just the relevant ones
            all_models (bool): if True, predict for each row using all models, not just the "model to use"
            prob_match_train (float): share of data-rows that are labeled
        '''

        if all_cols:

            phats = df.copy()

        else:

            cols_to_keep = [
                    'record_id_1', 'record_id_2', 'model_to_use',
                    'covered_pair', 'match_train_eligible', 'exactmatch', 'label']
            cols_to_keep = [col for col in cols_to_keep if col in df.columns]
            cols_to_keep = cols_to_keep + df.filter(regex='weight').columns.tolist()

            phats = df[cols_to_keep].copy()

        for model_name, mod in models.items():

            phat_col = f'{model_name}_{model_type}_phat'

            # initialize phat cols
            phats[phat_col] = np.NaN

            if oob:
                phats[phat_col] = mod.best_estimator_.named_steps['clf'].oob_decision_function_[:, MATCH_COL]

            else:

                if len(phats) == 0:
                    break

                if all_models:
                    phats[phat_col] = mod.predict_proba(df)[:, MATCH_COL]
                else:
                    to_predict_for = (phats.model_to_use == model_name)
                    phats.loc[to_predict_for, phat_col] = \
                            mod.predict_proba(df[to_predict_for])[:, MATCH_COL]

            if model_type == 'selection':
                weight_col = phat_col.replace('selection_phat', 'weight')
                weights = (prob_match_train + 1) / (phats[phat_col] + 1)
                phats[weight_col] = weights
                # P(s=1)/P(s=1|x), with smoothing

        return phats




from namematch.data_structures.schema import Schema
from namematch.data_structures.parameters import Parameters


def test_params_init(config_dict, default_params_dict):
        # test passing
        params = Parameters.init(
            config_dict,
            default_params_dict)

        schema = Schema.init(config_dict, params)

        # test wrong set_missing type
        config_dict['variables'][0]['drop'] = ','.join(config_dict['variables'][0]['drop'])
        params = Parameters.init(
                config_dict,
                default_params_dict)
        fine = 1
        try:
            schema = Schema.init(config_dict, params)
        except:
            fine = 0
        assert fine == 0



import argparse
import logging
import os
import yaml

from argcmdr import RootCommand, Command, main

logging.basicConfig(
    format="%(asctime)s - %(levelname)-8s %(message)s",
    datefmt="%d-%b-%y %H:%M:%S",
    level=logging.INFO
)


class NameMatch(RootCommand):
    """manage namematch"""

    def __init__(self, parser):
        parser.add_argument(
            "-c",
            "--config-file",
            help="configuration yaml file "
        )
        parser.add_argument(
            "-i",
            "--info-file",
            help="namematch info yaml file"
        )
        parser.add_argument(
            "--output-dir",
            default="output",
            help="output folder path (default: output)"
        )
        parser.add_argument(
            "--output-temp-dir",
            help="output temp folder path (default: <output_dir>/details)"
        )
        parser.add_argument(
            "--constraints-file",
            default=None,
            help="constraints file (optional)"
        )
        parser.add_argument(
            "-f",
            "--force",
            dest='force',
            action='store_true',
            help="force match to run even if outputs exist (default: False)"
        )
        parser.add_argument(
            "--trained-model-info-file",
            default='None',
            help="path to trained model from initial non-incremental run (required, only for incremental runs)"
        )
        parser.add_argument(
            "--existing-blocking-index-file",
            dest='og_blocking_index_file',
            default='None',
            help="path to existing blocking index from previous run (optional, only for incremental runs)"
        )
        parser.add_argument(
            "--enable-lprof",
            action='store_true',
            dest="enable_lprof",
            help="generate the line_profiler files for certain methods"
        )
        parser.add_argument(
            "--log-level",
            dest='log_level',
            default='INFO',
            help="logging level"
        )

    def get_namematcher(self):
        if bool(self.args.config_file) == bool(self.args.info_file):
            raise ValueError("Please only provide either config_file or info_file")


        from namematch.namematcher import NameMatcher

        if self.args.config_file:
            config = yaml.load(open(self.args.config_file, 'r'), Loader=yaml.FullLoader)

            if self.args.output_temp_dir:
                output_temp_dir = self.args.output_temp_dir
            else:
                output_temp_dir = os.path.join(self.args.output_dir, 'details')

            nm = NameMatcher(
                config=config,
                output_dir=self.args.output_dir,
                output_temp_dir=output_temp_dir,
                constraints=self.args.constraints_file,
                trained_model_info_file=self.args.trained_model_info_file,
                og_blocking_index_file=self.args.og_blocking_index_file,
                enable_lprof=self.args.enable_lprof,
                logging_level=self.args.log_level,
            )

        if self.args.info_file:
            logging.info("Loading from the stats file: {self.args.info_file}")
            nm = NameMatcher.load_namematcher(self.args.info_file)

        return nm


@NameMatch.register
class Run(Command):
    """Run all namematch steps"""
    def __call__(self, args):
        nm = self.root.get_namematcher()
        nm.run(force=self.root.args.force)


@NameMatch.register
class ProcessInputData(Command):
    """Process input data"""

    class Run(Command):
        def __init__(self, parser):
            parser.add_argument(
                '-f', '--force',
                dest='force',
                action='store_true',
                default=False,
                help="Process input data",
            )

        def __call__(self, args):
            nm = self.root.get_namematcher()
            nm.process_input_data.run(force=args.force)


@NameMatch.register
class GenerateMustLinks(Command):
    """Generate must links"""

    class Run(Command):
        def __init__(self, parser):
            parser.add_argument(
                '-f', '--force',
                dest='force',
                action='store_true',
                default=False,
                help="Genearte must links"
            )

        def __call__(self, args):
            nm = self.root.get_namematcher()
            nm.generate_must_links.run(force=args.force)


@NameMatch.register
class Block(Command):
    """Block"""

    class Run(Command):
        def __init__(self, parser):
            parser.add_argument(
                '-f', '--force',
                dest='force',
                action='store_true',
                default=False,
                help="Block"
            )

        def __call__(self, args):
            nm = self.root.get_namematcher()
            nm.block.run(force=args.force)


@NameMatch.register
class GenerateDataRows(Command):
    """Generate data rows"""

    class Run(Command):
        def __init__(self, parser):
            parser.add_argument(
                '-f', '--force',
                dest='force',
                action='store_true',
                default=False,
                help="Generate data rows"
            )

        def __call__(self, args):
            nm = self.root.get_namematcher()
            nm.generate_data_rows.run(force=args.force)


@NameMatch.register
class FitModel(Command):
    """Fit model"""

    class Run(Command):
        def __init__(self, parser):
            parser.add_argument(
                '-f', '--force',
                dest='force',
                action='store_true',
                default=False,
                help="Fit model"
            )

        def __call__(self, args):
            nm = self.root.get_namematcher()
            nm.fit_model.run(force=args.force)


@NameMatch.register
class Predict(Command):
    """Predict"""

    class Run(Command):
        def __init__(self, parser):
            parser.add_argument(
                '-f', '--force',
                dest='force',
                action='store_true',
                default=False,
                help="Predict"
            )

        def __call__(self, args):
            nm = self.root.get_namematcher()
            nm.predict.run(force=args.force)


@NameMatch.register
class Cluster(Command):
    """Cluster"""

    class Run(Command):
        def __init__(self, parser):
            parser.add_argument(
                '-f', '--force',
                dest='force',
                action='store_true',
                default=False,
                help="cluster"
            )

        def __call__(self, args):
            nm = self.root.get_namematcher()
            nm.cluster.run(force=args.force)


@NameMatch.register
class GenerateOutput(Command):
    """Generate output"""

    class Run(Command):
        def __init__(self, parser):
            parser.add_argument(
                '-f', '--force',
                dest='force',
                action='store_true',
                default=False,
                help="generate output"
            )

        def __call__(self, args):
            nm = self.root.get_namematcher()
            nm.generate_output.run(force=args.force)


@NameMatch.register
class GenerateReport(Command):
    """Generate report"""

    class Run(Command):
        def __init__(self, parser):
            parser.add_argument(
                '-f', '--force',
                dest='force',
                action='store_true',
                default=False,
                help="generate report"
            )

        def __call__(self, args):
            nm = self.root.get_namematcher()
            nm.generate_report.run(force=args.force)


def execute():
    main(NameMatch)


if __name__ == "__main__":
    main(NameMatch)



import logging
import os
from abc import ABC, abstractmethod
from pathlib import Path
import ruamel.yaml
yaml = ruamel.yaml.YAML()
yaml.indent(mapping=2, sequence=4, offset=2)

from ruamel.yaml.comments import CommentedMap

from namematch.utils.utils import reformat_dict, setup_logging, camel_to_snake
from namematch.data_structures.parameters import Parameters
from namematch.data_structures.schema import Schema

logger = logging.getLogger()

class NamematchBase(ABC):
    '''The base class for namematch steps

    All the inherited classes should have other input/output file paths for each step.

    Args:
        params (object): namematch's Parameter object
        schema (object): namematch's Schema object
        output_file (str): output file path

    '''
    def __init__(
        self,
        params: Parameters,
        schema: Schema,
        output_dir: str = None,
        nm_info_file=None,
        nm_metadata=None,
        profile_dir: str = None,
        next=None,
        enable_lprof=False,
    ):

        self.params = params
        self.schema = schema
        self.output_dir = output_dir
        self.stats_dict = CommentedMap()
        self.nm_info_file = nm_info_file
        self.nm_metadata = nm_metadata
        self.profile_dir = profile_dir
        self.next = next
        self.enable_lprof=enable_lprof

    @abstractmethod
    def main(self):
        '''Main method for each task class which is called in namematcher.NameMatcher'''
        raise NotImplementedError

    @property
    def output_files(self):
        return []

    @property
    def common_path(self):
        return Path(self.nm_info_file).parents[1]

    @property
    def yaml_ready_stats_dict(self):
        '''Change yaml format or style to create the yaml ready stats_dict'''
        stats_dict = reformat_dict(self.stats_dict)
        return {self.__class__.__name__: stats_dict}

    def check_output(self):
        all_output_files_existed = True
        if len(self.output_files) > 0:
            all_output_files_existed = []
            for output_file in self.output_files:
                if os.path.exists(output_file):
                    logger.info(f"Output '{os.path.relpath(output_file, self.common_path)}' exists!")
                    all_output_files_existed.append(True)
                else:
                    logger.debug(f"Output '{os.path.relpath(output_file, self.common_path)}' doesn't exist!")
                    all_output_files_existed.append(False)
            all_output_files_existed = all(all_output_files_existed)

        return all_output_files_existed

    def remove_output(self):
        if len(self.output_files) > 0:
            for f in self.output_files:
                if os.path.exists(f):
                    os.remove(f)
                    logger.info(f"Removing {os.path.relpath(f, self.common_path)} from {self.__class__.__name__}")

    def remove_downstream_output_and_stats(self):
        next_task = self.next
        nm_info = self.get_nm_info_with_synced_metadata()
        nm_stats = nm_info['stats']
        while next_task:
            # remove next task's output
            next_task.remove_output()
            task_name = next_task.__class__.__name__
            # remove next task's stats in stats file
            if nm_stats.get(task_name, None):
                del nm_stats[task_name]
            # empty next task's stats_dict
            next_task.stats_dict = CommentedMap()
            next_task = next_task.next

        with open(self.nm_info_file, 'w') as f:
            for key, value in nm_info.items():
                nm_info[key] = reformat_dict(value)
            yaml.dump(nm_info, f)

    def remove_current_task_from_stats_file(self):
        nm_info = self.get_nm_info_with_synced_metadata()
        nm_stats = nm_info['stats']
        if nm_stats.get(self.__class__.__name__, None):
            del nm_stats[self.__class__.__name__]

        nm_info['stats'] = nm_stats

        with open(self.nm_info_file, 'w') as f:
            for key, value in nm_info.items():
                nm_info[key] = reformat_dict(value)
            yaml.dump(nm_info, f)

    def run(self, force=False, write_stats_file=True):
        '''Run the Namematch Task

        Args:
            force (bool): whether to force the task to rerun even if the output exists

        '''
        try:

            if force:
                # remove current task's output
                self.remove_output()
                # remove current stats in stats file
                self.remove_current_task_from_stats_file()
                # empty current task's stats_dict
                self.stats_dict = CommentedMap()
                # remove all the downstream tasks' outputs and stats
                self.remove_downstream_output_and_stats()

            if not self.check_output():
                # remove current tasks' output
                self.remove_output()
                # remove current stats in stats file
                self.remove_current_task_from_stats_file()
                # remove current task's stats_dict
                self.stats_dict = CommentedMap()
                # if nm_stats file is not empty, remove downstream outputs and stats
                if self.get_nm_info_with_synced_metadata():
                    self.remove_downstream_output_and_stats()
                else:
                    if self.__class__.__name__ != "ProcessInputData":
                        raise Exception("Fatal Error. Please clean the output and rerun NameMatch.")
                logger.info(f"Running task: {self.__class__.__name__}")
                self.main()

                if write_stats_file:
                    logger.info(f"Writing stats for task: {self.__class__.__name__}")
                    try:
                        self.write_nm_info_to_yaml()
                    except Exception as e:
                        raise e

            else:
                logger.info(f"{self.__class__.__name__} does not run.")
                logger.info(f"The output of {self.__class__.__name__} exists. To force run, set `force` to True.")


        except KeyboardInterrupt as e:
            # remove current tasks' output
            self.remove_output()
            # remove current stats in stats file
            self.remove_current_task_from_stats_file()
            # remove current task's stats_dict
            self.stats_dict = CommentedMap()
            logger.critical(f"Keyboard Interrupt")
            logger.critical(f"Rolling back...")
            logger.critical(f"Next run will start from {self.__class__.__name__}")

            raise

    def get_nm_info_with_synced_metadata(self):
        '''Get the nm_info to sync metadata from the nm_info_file with the metadata from the namematch object.

        This is to update the metadata information from nm_info_file if one changes param/schem from the namematcher object.

        '''
        if os.path.exists(self.nm_info_file):
            with open(self.nm_info_file, 'r') as f:
                nm_info = yaml.load(f)
                nm_info['metadata'] = self.nm_metadata
        else:
            nm_info = CommentedMap()
            nm_info['metadata'] = self.nm_metadata
        return nm_info

    @property
    def nm_stats(self):
        return self.get_nm_info_with_synced_metadata()['stats']

    def write_nm_info_to_yaml(self):
        '''Write stats_dict to yaml file'''
        nm_metadata = self.get_nm_info_with_synced_metadata()

        nm_metadata['stats'].update(self.yaml_ready_stats_dict)
        with open(self.nm_info_file, 'w') as f:
            yaml.dump(nm_metadata, f)

    def write_line_profile_stats(self, profiler, output_unit=1e-6, stripzeros=False):
        current_lprof_file = camel_to_snake(self.__class__.__name__) + ".lprof"

        with open(os.path.join(self.profile_dir, current_lprof_file), 'w+') as f:
            profiler.print_stats(
                stream=f,
                output_unit=output_unit,
                stripzeros=stripzeros,
            )
            # clean up profiler
            profiler.functions = []
            profiler.code_map = {}





import tempfile
import os

from namematch.namematcher import NameMatcher

def test_namematcher(config_dict):
    with tempfile.TemporaryDirectory() as temp_dir:
        config_dict['num_workers'] = 1
        nm = NameMatcher(config=config_dict, output_dir=os.path.join(temp_dir, 'nm_output'))
        nm.process_input_data.run()
        nm.generate_must_links.run()
        nm.block.run()
        nm.generate_data_rows.run()
        nm.fit_model.run()
        nm.predict.run()
        nm.cluster.run()
        nm.generate_output.run()
        nm.generate_report.run()
        assert os.path.exists(os.path.join(temp_dir, 'nm_output', 'details', 'nm_info.yaml'))
        assert os.path.exists(os.path.join(temp_dir, 'nm_output', 'matching_report.html'))




import argparse
import errno
import logging
import os
import pickle
import shutil
import sys
import yaml

from datetime import datetime

from namematch.data_structures.parameters import *
from namematch.data_structures.data_file import *
from namematch.data_structures.variable import *
from namematch.utils import utils


class Schema():
    '''Class that houses the most essential instructions for how to complete the match: what
    data files to match, and which variables to use to do so.'''

    def __init__(self, data_files, variables):

        self.data_files = data_files
        self.variables = variables


    @classmethod
    def init(cls, config, params):
        '''Create and validate a DataFileList instance and a VariableList instance.

        Args:
            config (dict): dictionary with match parameter values
            params (dict): dictionary with processed match parameter values

        Returns:
            :mod:`namematch.data_structures.schema.Schema`: instance of the Schema class
        '''

        data_files = DataFileList.build(
                config['data_files'],
                config.get('existing_data_files', {}))
        data_files.validate()

        variables = VariableList.build(config['variables'], params)
        variables.validate(data_files)

        return cls(data_files, variables)


    @classmethod
    def load(cls, filepath):
        '''Load a Schema instance.

        Args:
            filepath (str): path to a yaml version of a Schema instance

        Returns:
            :mod:`namematch.data_structures.schema.Schema`: instance of the Schema class
        '''

        schema_dict = utils.load_yaml(filepath)

        data_files = DataFileList.load(schema_dict['data_files'])
        variables = VariableList.load(schema_dict['variables'])

        return cls(data_files, variables)

    @classmethod
    def load_from_dict(cls, schema_dict):

        data_files = DataFileList.load(schema_dict['data_files'])
        variables = VariableList.load(schema_dict['variables'])

        return cls(data_files, variables)

    def write(self, output_file):
        '''Write the Schema to a yaml file.

        Args:
            output_file (str): path to write schema dictionary
        '''

        utils.dump_yaml(utils.to_dict(self), output_file)


    # def copy(self):
    #     '''Create a deep copy of a Schema object.'''

    #     return Schema(
    #         self.data_files.copy(),
    #         self.variables.copy())



import numpy as np
import tempfile

from namematch.fit_model import define_necessary_models


def test_get_feature_info():
    pass


def test_fit_model():
    pass


def test_fit_models():
    pass


def test_define_necessary_models(data_rows_parquet_file):

    # fake data
    dr_file_list = [data_rows_parquet_file]
    missing_field = "first_name"

    with tempfile.TemporaryDirectory() as temp_output_dir:
        # missing
        model_info = define_necessary_models(dr_file_list, temp_output_dir, missing_field)
        assert 2 == len(model_info)
        assert "no_{}".format(missing_field) in model_info

        # not missing
        model_info = define_necessary_models(dr_file_list, temp_output_dir, None)
        assert 1 == len(model_info)


def test_get_match_train_eligible_flag():
    pass


def test_load_model_data():
    pass


def test_get_flipped0_potential_edges():
    pass





import logging
import os

from collections import defaultdict
import multiprocessing as mp
from math import ceil

import numpy as np
import pandas as pd
from namematch.name_probability import nm_prob

import pyarrow.parquet as pq
import pyarrow as pa

from namematch.utils.utils import *
from namematch.comparison_functions import *
from namematch.data_structures.schema import Schema
from namematch.data_structures.parameters import Parameters
from namematch.base import NamematchBase
from namematch.utils.profiler import Profiler

profile = Profiler()
logger = logging.getLogger()


class GenerateDataRows(NamematchBase):
    def __init__(
        self,
        params,
        schema,
        output_dir,
        all_names_file,
        candidate_pairs_file,
        *args,
        **kwargs
    ):
        super(GenerateDataRows, self).__init__(params, schema, *args, **kwargs)

        self.all_names_file = all_names_file
        self.candidate_pairs_file = candidate_pairs_file
        self.output_dir = output_dir

    @property
    def output_files(self):
        return [os.path.join(self.output_dir, f'data_rows_{i}.parquet') for i in range(self.params.num_workers)]

    # @log_runtime_and_memory
    def main(self, **kw):
        '''Take candidate pairs and merge on the all-names records (twice) to get a dataset at the
        record pair level. Compute distance metrics between the records in the pair -- these are the
        features for modeling.

        Args:
            params (Parameters object): contains parameter values
            schema (Schema object): contains match schema info (files to match, variables to use, etc.)
            all_names_file (str): path to the all-names file
            candidate_pairs_file (str): path to the candidate-pairs file
            output_dir (str): path to the data-rows dir
        '''

        if not os.path.exists(self.output_dir):
            os.mkdir(self.output_dir)

        an_columns = ['blockstring'] + self.schema.variables.get_an_column_names()
        table = pq.read_table(self.all_names_file)
        an = table.to_pandas()[an_columns]

        an = an[an.drop_from_nm == 0]
        an = an.drop_duplicates(['record_id'])

        logger.info('Building name probability object.')
        name_probs = self.generate_name_probabilities_object(
                an, self.params.first_name_column, self.params.last_name_column)

        logger.info('Generating data rows...')

        # remove previous data rows files if they exist
        for dr_file in os.listdir(self.output_dir):
            os.remove(os.path.join(self.output_dir, dr_file))

        table = pq.read_table(self.candidate_pairs_file)
        cp_df = table.to_pandas()[['blockstring_1', 'blockstring_2', 'covered_pair']]

        # get data rows
        end_points = get_endpoints(len(cp_df), self.params.num_workers)
        if self.params.parallelize:
            jobs = [mp.Process(
                        target=self.generate_data_row_files,
                        args=(self.params,
                              self.schema,
                              an,
                              cp_df.iloc[end_point[0]:end_point[1]],
                              name_probs,
                              end_point[0],
                              end_point[1],
                              os.path.join(self.output_dir, f'data_rows_{i}.parquet'))
                    ) for i, end_point in enumerate(end_points)]

            for job in jobs:
                job.start()
            for job in jobs:
                t = job.join()
            failure_occurred = sum([job.exitcode != 0 for job in jobs])
            if failure_occurred:
                logger.error(f"Error occurred in {failure_occurred} worker(s).")
                raise Exception(f"Error occurred in {failure_occurred} worker(s).")

        else:

            for i, end_point in enumerate(end_points):
                self.generate_data_row_files(
                        self.params, self.schema,
                        an, cp_df.iloc[end_point[0]:end_point[1]], name_probs,
                        end_point[0], end_point[1],
                        os.path.join(self.output_dir, f'data_rows_{i}.parquet'))

        if self.enable_lprof:
            self.write_line_profile_stats(profile.line_profiler)

    # @log_runtime_and_memory
    def generate_name_probabilities_object(self, an, fn_col=None, ln_col=None, **kw):
        '''The generate_name_probabilites function uses a list of names (from all_names
        file) and creates an object containing queryable probability
        information for each name.

        Args:
            an (pd.DataFrame): all-names, just the name columns
            fn_col (str): name of first name column
            ln_col (str): name of last name column

        Return:
            name probability object
        '''

        if (fn_col is None) or (ln_col is None):
            return None

        an['fn'] = an[fn_col].str.replace(' ', '')
        an['ln'] = an[ln_col].str.replace(' ', '')
        an['name_prob_str'] = '*' + an['fn'] + ' ' + an['ln'] + '*'

        np_object = nm_prob.NameProbability(name_list=an.name_prob_str.tolist())

        np_object.n_name_appearances_dict = an.groupby('name_prob_str').size().rank(pct=True, method='min').round(2).to_dict()
        np_object.n_firstname_appearances_dict = an.groupby('fn').size().rank(pct=True, method='min').round(2).to_dict()
        np_object.n_lastname_appearances_dict = an.groupby('ln').size().rank(pct=True, method='min').round(2).to_dict()

        return np_object

    # @log_runtime_and_memory
    def find_valid_training_records(self, an, an_match_criteria, **kw):

        an['meets_match_criteria'] = 1
        for col, accepted_values in an_match_criteria.items():
            if not isinstance(accepted_values, list):
                accepted_values = [accepted_values]
            accepted_values = [str(av) for av in accepted_values]
            an['meets_match_criteria'] = (an['meets_match_criteria']) & (an[col].isin(accepted_values))

        return an['meets_match_criteria']

    @profile
    def generate_actual_data_rows(self, params, schema, sbs_df, np_object, first_iter):
        '''Create modeling dataframe by comparing each variable (via numerous distance metrics).

        Args:
            params (Parameters object): contains matching parameters
            schema (Schema object): contains matching schema (data files and variables)
            sbs_df (pd.DatFrame): side-by-side table (record pair level, with info from both an records)

                ==============================   =======================================================
                record_id (_1, _2)               unique record identifier
                blockstring (_1, _2)             concatenated version of blocking columns (sep by ::)
                file_type (_1, _2)               either "new" or "existing"
                candidate_pair_ix                index from candidate-pairs list
                covered_pair                     flag, 1 if blockstring pair passed blocking 0 otherwise
                <fields for matching> (_1, _2)   both for the matching model and for constraint checking
                ==============================   =======================================================
            np_object (nm_prob.NameProbability object): contains information about name probabilities

        Returns:
            pd.DataFrame: chunk of the data-rows file

            =====================   =======================================================
            dr_id                   unique record pair identifier (record_id_1__record_id_2)
            record_id (_1, _2)      unique record identifiers
            <distance metrics>      how similar are the different matching fields between recrods
            label                   "1" if the records refer to the same person, "0" if not, "" otherwise
            =====================   =======================================================
        '''

        sbs_df = sbs_df.copy()

        uid_cols = schema.variables.get_variables_where(attr='compare_type', attr_value='UniqueID')

        # don't compare a record to itself
        sbs_df = sbs_df[sbs_df.record_id_1 != sbs_df.record_id_2]

        # if blockstrings are same, don't want to compare A to B then B to A
        sbs_df = sbs_df[
                (sbs_df.blockstring_1 != sbs_df.blockstring_2) |
                (sbs_df.record_id_1 < sbs_df.record_id_2)]

        if params.incremental:

            # dont compare if both are from existing (this would work outside of
            # if statement, but would always be true -- do it here to save time)
            sbs_df = sbs_df[
                    (sbs_df.file_type_1 == 'new') |
                    (sbs_df.file_type_2 == 'new')]

        if params.drop_mixed_label_pairs:
            for uid_col in uid_cols:
                sbs_df = sbs_df[(sbs_df[f"{uid_col}_1"] == '') == (sbs_df[f"{uid_col}_2"] == '')]

        if len(sbs_df) == 0:
            return None

        non_feature_cols = ['candidate_pair_ix', 'record_id_1', 'record_id_2', 'covered_pair']

        data_rows_df = sbs_df[non_feature_cols].copy()

        data_rows_df.reset_index(drop=True, inplace=True)
        sbs_df.reset_index(drop=True, inplace=True)

        if np_object is not None:

            # get name probabilities (both as feature and for knowing which
            # name to switch if a switch is needed)
            sbs_df = get_name_probabilities(sbs_df, np_object,
                    params.first_name_column, params.last_name_column)

            # determine if name switch need to happen
            sbs_df = try_switch_first_last_name(sbs_df,
                    params.first_name_column, params.last_name_column)

        if any(data_rows_df.index != sbs_df.index):
            logger.error('Index mismatch when creating data rows.')
            raise ValueError

        # add name probability columns to data_rows_df
        try:
            #data_rows_df['prob_name1'] = sbs_df['prob_name_1']
            #data_rows_df.loc[(sbs_df.switched_name == 1), 'prob_name1'] = sbs_df.prob_rev_name_1
            #data_rows_df['prob_name2'] = sbs_df['prob_name_2']
            #data_rows_df.loc[(sbs_df.switched_name == 2), 'prob_name2'] = sbs_df['prob_rev_name_2']
            #data_rows_df['prob_same_name'] = sbs_df['prob_same_name']
            #data_rows_df.loc[sbs_df.switched_name == 1, 'prob_same_name'] = sbs_df['prob_same_name_rev_1']
            #data_rows_df.loc[sbs_df.switched_name == 2, 'prob_same_name'] = sbs_df['prob_same_name_rev_2']
            #data_rows_df['max_prob_name'] = data_rows_df[['prob_name1', 'prob_name2']].max(axis=1)
            #data_rows_df['count_pctl_name_1'] = sbs_df['count_pctl_name_1']
            #data_rows_df['count_pctl_name_2'] = sbs_df['count_pctl_name_2']
            #data_rows_df['max_count_pctl_name'] = data_rows_df[['count_pctl_name_1', 'count_pctl_name_2']].max(axis=1)
            data_rows_df['diff_count_pctl_name'] = (sbs_df['count_pctl_name_1'] - sbs_df['count_pctl_name_2']).abs()
            data_rows_df['max_count_pctl_name'] = sbs_df[['count_pctl_name_1', 'count_pctl_name_2']].max(axis=1)
            data_rows_df['diff_count_pctl_fn'] = (sbs_df['count_pctl_fn_1'] - sbs_df['count_pctl_fn_2']).abs()
            data_rows_df['max_count_pctl_fn'] = sbs_df[['count_pctl_fn_1', 'count_pctl_fn_2']].max(axis=1)
            data_rows_df['diff_count_pctl_ln'] = (sbs_df['count_pctl_ln_1'] - sbs_df['count_pctl_ln_2']).abs()
            data_rows_df['max_count_pctl_ln'] = sbs_df[['count_pctl_ln_1', 'count_pctl_ln_2']].max(axis=1)

        except:
            if first_iter:
                logger.info('No name probability features generated.')

        for variable in schema.variables.varlist:

            if variable.compare_type == 'String':
                features_df = compare_strings(sbs_df, variable.name)
            elif variable.compare_type == 'Numeric':
                features_df = compare_numbers(sbs_df, variable.name)
            elif variable.compare_type == 'Categorical':
                features_df = compare_categories(sbs_df, variable.name)
            elif variable.compare_type == 'Date':
                features_df = compare_dates(sbs_df, variable.name)
            elif variable.compare_type == 'Geography':
                features_df = compare_geographies(sbs_df, variable.name)
            elif variable.compare_type == 'Address':
                features_df_1 = compare_strings(sbs_df, 'address_street_number')
                features_df_2 = compare_strings(sbs_df, 'address_street_name')
                features_df_3 = compare_categories(sbs_df, 'address_street_type')
                features_df = pd.concat([features_df_1, features_df_2, features_df_3], axis=1)
            elif variable.compare_type == "LastName":
                ##features_df_1 = compare_strings(sbs_df, variable.name)
                ##features_df_2 = compare_last_name(sbs_df, variable.name)
                ##features_df = pd.concat([features_df_1, features_df_2], axis=1)
                # TODO add back in some sort of JR, SR check
                features_df = compare_strings(sbs_df, variable.name)
            else:
                features_df = None

            if features_df is not None:
                data_rows_df = pd.concat([data_rows_df, features_df], axis=1)

        data_rows_df.columns = ['var_' +  colname if colname not in non_feature_cols else colname
                for colname in data_rows_df.columns]

        try:
            data_rows_df['exactmatch'] = 1
            for exact_match_col in params.exact_match_variables:
                data_rows_df['exactmatch'] = \
                        ((data_rows_df.exactmatch) &
                        (sbs_df[f'{exact_match_col}_1'] == sbs_df[f'{exact_match_col}_2']) &
                        (sbs_df[f'{exact_match_col}_1'] != '')).astype(int)
            data_rows_df['var_exact_match'] = data_rows_df.exactmatch.copy()
            for neg_var in params.negate_exact_match_variables:
                data_rows_df.loc[(sbs_df[f'{neg_var}_1'] != sbs_df[f'{neg_var}_2']) &
                                (sbs_df[f'{neg_var}_1'] != '') &
                                (sbs_df[f'{neg_var}_2'] != ''), 'var_exact_match'] = 0
        except:
            pass

        data_rows_df['label'] = generate_label(
                sbs_df,
                uid_cols,
                params.leven_thresh)

        if data_rows_df is not None and len(data_rows_df) > 0:
            data_rows_df['dr_id'] = data_rows_df.record_id_1 + '__' + data_rows_df.record_id_2
            data_rows_df.set_index('dr_id', inplace=True)

        return data_rows_df

    # @log_runtime_and_memory
    @profile
    def generate_data_row_files(
                self, params, schema, an, cp_df, name_probs,
                start_ix_worker, end_ix_worker, dr_file, **kw):
        '''The get_data_row_files function is run in parallel to generate the data needed for
        the random forest; it performs the merge between candidate pairs and all-names and
        calls the function that calculates distance metrics.

        Args:
            params (Parameters object): contains matching parameters
            schema (Schema object): contains matching schema (data files and variables)
            an (pd.DatFrame): all-names table (one row per input record)

                =====================   =======================================================
                record_id               unique record identifier
                file_type               either "new" or "existing"
                <fields for matching>   both for the matching model and for constraint checking
                <raw name fields>       pre-cleaning version of first and last name
                blockstring             concatenated version of blocking columns (sep by ::)
                drop_from_nm            flag, 1 if met any "to drop" criteria 0 otherwise
                =====================   =======================================================

            cp_df (pd.DataFrame): candidate-pairs list

                ======================   =======================================================
                blockstring_1            concatenated version of blocking columns for first element in pair (sep by ::)
                blockstring_2            concatenated version of blocking columns for second element in pair (sep by ::)
                covered_pair             flag; 1 for pairs that made it through blocking, 0 otherwise; all 1s here
                ======================   =======================================================

            name_probs (nm_prob.NameProbability object): contains information about name probabilities
            start_ix_worker (int): starting index of the candidate-pairs chunk to read in this thread
            end_ix_worker (int): end index of the candidate-pairs chunk to read in this thread
            dr_file (str): path to data-rows file to write (one for each worker thread)
        '''
        try:
            cp_df = cp_df.copy()

            # sqlite replacement
            an_bs = an[['blockstring']].copy()
            an_bs['i'] = np.arange(len(an_bs))
            an_ix_map = defaultdict(lambda: [])
            for tup in an_bs.itertuples():
                an_ix_map[tup.blockstring].append(tup.i)

            # for each candidate pair, get the records associated with each name and compute
            # distances between features like date, first_name, dob, etc.

            cp_df['candidate_pair_ix'] = cp_df.index

            self.num_batches = ceil(len(cp_df) / self.params.data_rows_batch_size)
            start_ix_cp = 0
            while start_ix_cp < len(cp_df):
                if (start_ix_worker == 0) and (params.verbose is not None) and (start_ix_cp % params.verbose == 0):
                    logger.info(f"   Generated features for {start_ix_cp * params.num_workers} "
                                f"of ~{len(cp_df) * params.num_workers} pairs of blockstrings.")

                end_ix_cp = min(start_ix_cp + self.params.data_rows_batch_size, len(cp_df))

                relevant_cp = cp_df.iloc[start_ix_cp : end_ix_cp].copy()
                relevant_blockstrings = pd.concat([relevant_cp.blockstring_1,
                                                   relevant_cp.blockstring_2]).unique().tolist()
                relevant_bs_ix = [an_ix_map[bs] for bs in relevant_blockstrings]
                relevant_bs_ix = [item for sublist in relevant_bs_ix for item in sublist]

                # get relevant records for this batch
                relevant_records = an.iloc[relevant_bs_ix].copy()
                relevant_records.set_index('blockstring', inplace=True)

                # expand the candidate pairs from a blockstring level to a record level
                side_by_side_df = pd.merge(
                    relevant_cp[['blockstring_1', 'blockstring_2', 'candidate_pair_ix', 'covered_pair']],
                    relevant_records.copy(),
                    left_on='blockstring_1', right_index=True).reset_index(drop=True)
                side_by_side_df = pd.merge(
                    side_by_side_df,
                    relevant_records.copy(),
                    left_on='blockstring_2', right_index=True, suffixes=['_1', '_2'])

                data_rows_df = self.generate_actual_data_rows(
                        params, schema, side_by_side_df, name_probs, first_iter=(start_ix_cp == 0))

                if data_rows_df is None:
                    start_ix_cp += self.params.data_rows_batch_size
                    continue

                # outcome for selection model
                data_rows_df['labeled_data'] = (data_rows_df.label != '').astype(int)

                table = pa.Table.from_pandas(data_rows_df)
                if start_ix_cp == 0:
                    pqwriter = pq.ParquetWriter(dr_file, table.schema)

                pqwriter.write_table(table)

                start_ix_cp += self.params.data_rows_batch_size

            if pqwriter:
                pqwriter.close()

        except Exception as e:
            os.remove(dr_file)
            logger.error(f"{dr_file} failed.")
            raise e




import logging
import numpy as np
import os
import pandas as pd
import yaml

import pyarrow as pa
import pyarrow.parquet as pq

from collections import defaultdict
from datetime import datetime
from streetaddress import StreetAddressParser

from namematch.base import NamematchBase
from namematch.data_structures.schema import Schema
from namematch.data_structures.parameters import Parameters
from namematch.utils.utils import (
    build_blockstring,
    clean_nn_string,
    create_nm_record_id,
    log_runtime_and_memory,
    load_yaml,
)
from namematch.utils.profiler import Profiler

profile = Profiler()

logger = logging.getLogger()

class ProcessInputData(NamematchBase):
    def __init__(
        self,
        params,
        schema,
        all_names_file="all_names.parquet",
        *args,
        **kwargs
    ):
        super(ProcessInputData, self).__init__(params, schema, *args, **kwargs)

        self.all_names_file = all_names_file

    @property
    def output_files(self):
        return [
            self.all_names_file
        ]

    # @log_runtime_and_memory
    def main(self, **kw):
        '''Follow the instructions in the schema and params objects to build the all-names file from
        the raw input file(s).

        Args:
            params (Parameters object): contains parameter values
            schema (Schema object): contains match schema info (files to match, variables to use, etc.)
            all_names_file (str): path to the all-names file
        '''
        self.stats_dict["start"] = datetime.now().strftime("%m/%d/%Y, %H:%M:%S")
        self.stats_dict.yaml_add_eol_comment("process_input_data", "start")

        if self.params.incremental:
            logger.info("This is an incremental run.")
            self.stats_dict["run_type"] = "incremental"

        else:
            logger.info("This is not an incremental run.")
            self.stats_dict["run_type"] = "from scratch"

        logger.trace("Start: process_input_data")

        data_file_list = self.schema.data_files.get_all_data_files()

        # store categorical variables for report
        cat_vars = [v.name for v in self.schema.variables.varlist if v.compare_type == "Categorical"]
        self.stats_dict["categorical_variables"] = cat_vars

        n_an_rows = 0
        n_valid_an_rows = 0
        for j, data_file in enumerate(data_file_list):

            logger.info(f'Reading {data_file.nickname} data.')

            raw_columns_data = self.schema.variables.get_columns_to_read(data_file)

            for i, df in enumerate(pd.read_csv(
                    data_file.filepath,
                    chunksize=self.params.input_data_batch_size,
                    usecols=raw_columns_data,
                    dtype=object,
                    encoding="ISO-8859-1", 
                    na_filter=False)):

                if len(df) == 0:
                    break

                if self.params.verbose != None:
                    logger.info(f'  Processing: read {(i+1) * self.params.input_data_batch_size} rows.')

                df = self.process_data(df, self.schema.variables, data_file, self.params)

                first_iteration = (os.path.exists(self.all_names_file) == False)
                if len(df) > 0:
                    n_an_rows += len(df)
                    n_valid_an_rows += sum(df.drop_from_nm == 0)

                    if i == 0 and j == 0:
                        table = pa.Table.from_pandas(df)
                        pqwriter = pq.ParquetWriter(self.all_names_file, table.schema)
                    else:
                        table = pa.Table.from_pandas(df, schema=table.schema)

                    pqwriter.write_table(table)

                    # NOTE: append is okay even for first iteration because always
                    #       deleting existing files above

            logger.info(f'Done writing {data_file.nickname} data.')

        if pqwriter:
            pqwriter.close()

        logger.info(f"Number of input records: {n_an_rows}")
        logger.info(f"Number of valid input records: {n_valid_an_rows}")
        self.stats_dict["n_an"] = n_an_rows
        self.stats_dict["n_valid_an"] = n_valid_an_rows
        logger.trace("End: process_input_data")

        if self.enable_lprof:
            self.write_line_profile_stats(profile.line_profiler)

    @profile
    def process_geo_column(self, df, variable):
        '''Take dataframe of geographic data (either in "lat,lon" format or in
        "lat", "lon" format) and ensure it has just one column.

        Args:
            df (pd.DataFrame): df of address input data (columns are strings)
            variable (Variable object): contains naming info for new geo column

        Returns:
            pd.Dataframe: DataFrame of clean geographic information for all_names file
        '''

        df = df.copy()
        variable = variable.copy()

        ncols = len(df.columns)

        if ncols == 2:
            # concatenate two values (either x and y or lat and long)
            # if either column is empty, set the new column to be empty
            col1 = df.columns.tolist()[0]
            col2 = df.columns.tolist()[1]
            df[variable.name] = ''
            df.loc[(df[col1] != '') & (df[col2] != ''), variable.name] = \
                    df[col1].astype(str).str.strip() + ',' + df[col2].astype(str).str.strip()

        elif ncols == 1:
            # must already have comma in row (e.g. "12254231,95236928")

            df[variable.name] = df[df.columns.tolist()[0]]
            df['no_comma'] = ((df[variable.name] != '') & \
                    (df[variable.name].str.contains(',') == False)).astype(int)
            if df.no_comma.sum() > 0:
                logger.trace(f'Setting {variable.name} variable to missing for '
                             f'{df.no_comma.sum()} rows (misspecified).')
            df.loc[df.no_comma == 1, variable.name] = ''

        else:
            logger.error(f"Variables of compare_type Geography take "
                         f"exactly 0, 1, or 2 column names.")
            raise ValueError

        return df[[variable.name]]


    @profile
    def parse_address(self, address):
        '''Parse an address string into distinct parts.

        Args:
            address (str): string of full address (e.g. 54 East 18th Rd.)

        Returns:
            tuple: (address number, street name, street suffix)
        '''

        add_map = defaultdict(str, {
            "ave": "avenue", "avenue": "avenue",
            "blvd": "boulevard", "boulevard": "boulevard",
            "ctr": "center", "center": "center",
            "cir": "circle", "circle": "circle",
            "ct": "court", "court": "court",
            "cv": "cove", "cove": "cove",
            "dr": "drive", "drive": "drive",
            "expy": "expressway", "expressway": "expressway",
            "hts": "heights", "heights": "heights",
            "hwy": "highway", "highway": "highway",
            "jct": "junction", "junction": "junction",
            "ln": "lane", "lane": "lane",
            "lp": "loop", "loop": "loop",
            "pkwy": "parkway", "parkway": "parkway",
            "pl": "place", "place": "place",
            "rd": "road", "road": "road",
            "sq": "square", "square": "square",
            "st": "street", "street": "street"
        })
        # TODO need ter for terrace (but would have to change that in the street address parser...)
        # TODO need park (but would have to change that in the street address parser...)
        # TODO addresses with no street type (can lead to city name getting included in street name)
        # TODO should I not include street suffix as feature? don't really want two things to get
        #      "points" because both live on an avenue...

        address_parser = StreetAddressParser()
        address = address.lower()
        if address == 'redacted' or address == 'redact':
            address = ''
        ap = address_parser.parse(address)
        ap['house'] = '' if not ap['house'] else ap['house']
        ap['street_name'] = '' if not ap['street_name'] else ap['street_name']
        ap['street_type'] = '' if not ap['street_type'] else add_map[ap['street_type']]

        return ap['house'], ap['street_name'], ap['street_type']


    @profile
    def process_address_column(self, df, logger=None):
        '''Take dataframe of address data (either in "123 Main St." format or
        "123", "Main", "St." format, order matters) and parse as needed to produce
        three clean columns: street number, street name, and street type.

        Args:
            df (pd.DataFrame): df of address input data

        Returns:
            pd.DataFrame: Dataframe of clean address information for all_names file
        '''

        df = df.copy()

        ncols = len(df.columns)
        addr_cols = ['address_street_number', 'address_street_name', 'address_street_type']

        if ncols == 3:
            # simply rename columns
            df.columns = addr_cols

        elif ncols == 1:
            # parse address
            parsed_df = np.vectorize(self.parse_address)(df[df.columns.tolist()[0]].values)
            parsed_df = list(zip(*parsed_df))
            df[addr_cols] = pd.DataFrame(parsed_df, index=df.index)

        else:
            logger.error("Variables of compare_type Address take exactly 0, 1, or 3 column names.")
            raise ValueError

        return df[addr_cols]


    @profile
    def process_check(self, s, variable):
        '''Check the validity of the values in a given all-names column (according to the data type
        and config instructions) and set the series name correctly.

        Args:
            s (pd.Series): series to process (will be an all-names column)
            variable (Variable object): contains info on how to validate data in series

        Returns:
            pd.Series: Processed series
        '''

        s = s.copy()
        variable = variable.copy()

        if variable.compare_type != 'Address':
            s.name = variable.name

        if variable.check == 'Numeric':
            not_numeric = ((s != '') & (pd.to_numeric(s, errors='coerce')).isnull())
            if not_numeric.sum() > 0:
                logger.trace(f'Setting {s.name} variable to missing for {not_numeric.sum()} '
                             f'rows (fails check).')
            s.loc[not_numeric] = ''

        elif variable.check.startswith('Date'):
            before = s.isnull().sum()
            date_format = variable.check.split('-', 1)[1].strip()
            s = pd.to_datetime(s, format=date_format, errors="coerce")
            after = s.isnull().sum()
            if (before - after) > 0:
                logger.trace(f'Setting {s.name} variable to missing for '
                             f'{before - after} rows (fails check).')
            # convert to string in python date format: e.g. 2014-05-31
            s = s.dt.strftime('%Y-%m-%d')
            s = s.replace('NaT', '')
            s = s.replace(np.NaN, '')

        elif variable.compare_type == 'Categorical' and variable.check != '':
            options = variable.check.split(',')
            not_in_options = s[s.isin(options) == False].index.tolist()
            if len(not_in_options) > 0:
                logger.trace(f'Setting {s.name} variable to missing for '
                             f'{len(not_in_options)} rows (fails check).')
            s.loc[not_in_options] = ''

        else:
            try:
                s = s.str.strip()
                s = s.str.upper()
            except:
                raise

        return s

    @profile
    def process_data(self, df, variables, data_file, params):
        '''Read in part of an input file and process it according to the config in order
        to create part of the all-names file.

        Args:
            df (pd.DataFrame): chunk of an input data file
            variables (VariableList object): contains info about the fields for matching (from config)
            data_file (DataFile object): contains info about the input data set
            params (dict): dictionary of param values

        Returns:
            pd.DataFrame: a chunk of the all-names table (one row per input record)

            =====================   =======================================================
            record_id               unique record identifier
            file_type               either "new" or "existing"
            <fields for matching>   both for the matching model and for constraint checking
            <raw name fields>       pre-cleaning version of first and last name
            blockstring             concatenated version of blocking columns (sep by ::)
            drop_from_nm            flag, 1 if met any "to drop" criteria 0 otherwise
            =====================   =======================================================
        '''

        df = df.copy()
        variables = variables.copy()
        data_file = data_file.copy()
        params = params.copy()

        if data_file.use_record_id_as_is:
            df['record_id'] = df[data_file.record_id_col]
        else:
            df['record_id'] = create_nm_record_id(
                    data_file.nickname,
                    df[data_file.record_id_col])
        df.set_index('record_id', inplace=True)

        # create a new data frame to store all the processed variables
        an = pd.DataFrame(index=df.index)
        an['file_type'] = data_file.file_type
        an['dataset'] = data_file.nickname

        # track the records that will need to be dropped-- don't drop until the end
        all_records_to_drop = []

        for variable in variables.varlist:

            relevant_columns = variable.get_columns_to_read(data_file.nickname)

            # handle case where data file doesn't contain column
            if (variable.colname_dict[data_file.nickname] == ''):
                if (variable.compare_type == "Address"):
                    an["address_street_number"] = ""
                    an["address_street_name"] = ""
                    an["address_street_type"] = ""
                else:
                    an[variable.name] = ''
                continue

            # separate or combine columns as needed
            if (variable.compare_type == 'Geography'):
                relevant_df = self.process_geo_column(df[relevant_columns], variable)
            elif (variable.compare_type == 'Address'):
                relevant_df = self.process_address_column(df[relevant_columns])
            else:
                # if the compare_type isn't Geography or Address, there should only be one column
                if (len(relevant_columns) > 1):
                    logger.error(f'Only variables of compare_type "Geography" and "Address" '
                                 f'can take multiple column names.')
                    raise ValueError
                relevant_df = df[[relevant_columns[0]]]

            # NOTE: the only variable that should cause relevant_df to
            # have more than one column is Address
            for col in relevant_df:

                s = relevant_df[col].copy()
                raw_s = s.copy()

                # clean up last names
                if variable.name == params.last_name_column:
                    s = s.str.replace("MC ", "MC")
                    s = s.str.replace("-", " ")

                if len(variable.set_missing_list) > 0:
                    s = process_set_missing(s, variable.set_missing_list)

                s = self.process_check(s, variable)

                # clean up strings for blocking
                if variable.name in params.blocking_scheme['cosine_distance']['variables']:
                    an[f'tmp_raw__{s.name}'] = raw_s
                    s = s.apply(clean_nn_string)

                # NOTE: always drop last (in case other processing causes
                # value to go into the drop_list)
                if len(variable.drop_list) > 0:
                    records_to_drop = process_drop(s, variable.drop_list)
                    all_records_to_drop.extend(records_to_drop)

                an[s.name] = s

        # clean up the index
        an.reset_index(inplace=True)

        # process auto_drops
        auto_drop_records = process_auto_drops(an, all_records_to_drop, params.auto_drop_logic)
        all_records_to_drop = all_records_to_drop + auto_drop_records

        an['drop_from_nm'] = an.record_id.isin(all_records_to_drop).astype(int)
        an['blockstring'] = build_blockstring(an, params.blocking_scheme)

        return an

def process_set_missing(s, set_missing_list):
    '''Set values in a series to missing as needed.

    Args:
        s (pd.Series): strings to process
        set_missing_list (list): list of strings that are disallowed

    Returns:
        pd.Series: Processed series
    '''

    s = s.copy()
    missing = s.isin(set_missing_list)
    if (missing.sum() > 0):
        logger.trace(f'Setting {s.name} variable to missing for {missing.sum()} '
                     f'rows (in set_missing).')
    s.loc[missing] = ''

    return s


def process_drop(s, drop_list):
    '''Get the records in a series that have invalid values.

    Args:
        s (pd.Series): series being processed
        drop_list (list of str): invalid values

    Returns:
        list: Indices of records that are not valid
    '''

    s = s.copy()
    records_to_drop = s[s.isin(drop_list)].index.tolist()
    if (len(records_to_drop) > 0):
        logger.trace(f'Adding {len(records_to_drop)} rows to drop list because '
                     f'of the {s.name} variable.')

    return records_to_drop


def process_auto_drops(an, existing_drop_list, drop_logic):
    '''Get the records in all-names that have invalid values due to combination
    of multiple columns (based on logic in the private config).

    Args:
        an (pd.DataFrame): all-names chunk being processed
        existing_drop_list (list of str): records already known to be invalid
        drop_logic (list of dicts): logic for what makes a record invalid

    Returns:
        list: Indices of records that are not valid
    '''

    an = an.copy()
    existing_drop_list = existing_drop_list[:]
    drop_logic = drop_logic[:]

    an = an[an.record_id.isin(existing_drop_list) == False]

    to_drop_set = set()

    for invalid_reason in drop_logic:

        an['invalid'] = 1
        for col, value in invalid_reason.items():
            an.loc[an[col] != value, 'invalid'] = 0

        to_drop_set.update(an[an.invalid == 1].record_id.tolist())

    auto_drop_records = list(to_drop_set)

    return auto_drop_records






import os
import logging
import tempfile
import papermill as pm
from nbconvert import HTMLExporter
from namematch.base import NamematchBase
from namematch.data_structures.parameters import Parameters
from namematch.data_structures.schema import Schema

class IgnoreBlackWarning(logging.Filter):
    def filter(self, record):
        return 'Black is not installed' not in record.msg


logging.getLogger("papermill.translators").addFilter(IgnoreBlackWarning())
logger = logging.getLogger()


class GenerateReport(NamematchBase):
    '''
    params (Parameters object): contains parameter values
    schema (Schema object): contains match schema info (files to match, variables to use, etc.)
    report_file (str): full path of the report html file
    '''
    def __init__(self, params, schema, report_file, *args, **kwargs):
        super(GenerateReport, self).__init__(params, schema, *args, **kwargs)
        self.report_file = report_file

    @property
    def output_files(self):
        return [self.report_file]


    def main(self, **kw):
        with tempfile.TemporaryDirectory() as temp_dir:
            temp_output_notebook = os.path.join(temp_dir, 'matching_report_output.ipynb')
            # Execute notebook as a temp notebook
            pm.execute_notebook(
                os.path.join(os.path.dirname(__file__), 'matching_report.ipynb'),
                temp_output_notebook,
                parameters=dict(nm_info_path=self.nm_info_file),
                kernel_name='python3',
                progress_bar=False
            )
            # Export to html
            logger.info("Exporting to html")
            html_exporter = HTMLExporter()
            html_exporter.exclude_input = True
            html_exporter.exclude_input_prompt = True
            html_exporter.exclude_output_prompt = True
            html_data, resources = html_exporter.from_filename(temp_output_notebook)
            with open(self.report_file, 'w', encoding='utf-8', errors='ignore') as f:
                f.write(html_data)

            logger.info(f"The matching report is created at {self.report_file}")



from namematch.utils.utils import clean_nn_string, build_blockstring


def test_create_nm_record_id():
    pass


def test_clean_nn_string():

    # test JR and .
    assert clean_nn_string("JOHN SMITH JR.") == 'JOHN SMITH'

     # test STRIP, III, and -
    assert clean_nn_string(" JOHN  SM-ITH III") == 'JOHN SMITH'

    # test strip non A-Z (uppercase)
    assert clean_nn_string("JOHN Smith") == 'JOHN S'

    # test that uppercase isn't handled
    assert clean_nn_string("JOHN smith") != 'JOHN SMITH'


def test_build_blockstring(an_df):
    scheme = {
        "cosine_distance": {
            "variables": ["first_name", "last_name"],
	    "relative_weight_of_second_variable": 1.2
        },
        "edit_distance": {"variable": "dob"}
    }

    # get blockstrings
    blockstring_col = build_blockstring(an_df, scheme)

    # test
    assert an_df.shape[0] == len(blockstring_col)
    expected_num_cols = len(scheme['cosine_distance']['variables'] + [scheme['edit_distance']['variable']])
    num_cols = len(blockstring_col[0].split("::"))
    assert expected_num_cols == num_cols


def test_determine_model_to_use():
    pass


def test_get_nn_string_from_blockstring():
    pass


def test_get_ed_string_from_blockstring():
    pass


def test_get_endpoints():
    pass


